"""Module to do differentiable rendering of 2D and 3D scenes."""

import copy
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, overload, Tuple, Union
import warnings

import numpy as np
from typing_extensions import Literal

from deodr.triangulated_mesh import ColoredTriMesh

from . import differentiable_renderer_cython  # type: ignore


@dataclass
class Scene2DBase:
    """Class representing the structure representing the 2.5
    scene expected by the C++ code
    """

    faces: np.ndarray
    faces_uv: np.ndarray
    ij: np.ndarray
    depths: np.ndarray
    textured: np.ndarray
    uv: np.ndarray
    shade: np.ndarray
    colors: np.ndarray
    shaded: np.ndarray
    edgeflags: np.ndarray
    height: int
    width: int
    nb_colors: int
    texture: np.ndarray
    background_image: Optional[np.ndarray] = None
    background_color: Optional[np.ndarray] = None
    uv_b: Optional[np.ndarray] = None
    ij_b: Optional[np.ndarray] = None
    shade_b: Optional[np.ndarray] = None
    colors_b: Optional[np.ndarray] = None
    texture_b: Optional[np.ndarray] = None
    clockwise: bool = False
    backface_culling: bool = True
    strict_edge: bool = True
    perspective_correct: bool = False
    integer_pixel_centers: bool = True


def renderScene(
    scene: Scene2DBase,
    sigma: float,
    image: np.ndarray,
    z_buffer: np.ndarray,
    antialiase_error: bool = False,
    obs: Optional[np.ndarray] = None,
    err_buffer: Optional[np.ndarray] = None,
    check_valid: bool = True,
) -> None:
    if check_valid:
        # doing checks here as it seems the debugger in not able to find the pyx file
        # when installed from a wheel. this also make interactive debugging easier
        # for the library user

        assert image is not None
        assert z_buffer is not None
        height = image.shape[0]
        width = image.shape[1]
        nb_colors = image.shape[2]

        nb_triangles = scene.faces.shape[0]
        assert nb_triangles == scene.faces_uv.shape[0]
        nb_vertices = scene.depths.shape[0]
        nb_vertices_uv = scene.uv.shape[0]

        assert scene.faces.dtype == np.uint32
        assert np.all(scene.faces < nb_vertices)
        assert np.all(scene.faces_uv < nb_vertices_uv)

        assert scene.colors.ndim == 2
        assert scene.uv.ndim == 2
        assert scene.ij.ndim == 2
        assert scene.shade.ndim == 1
        assert scene.edgeflags.ndim == 2
        assert scene.textured.ndim == 1
        assert scene.shaded.ndim == 1
        assert scene.uv.shape[1] == 2
        assert scene.ij.shape[0] == nb_vertices
        assert scene.ij.shape[1] == 2
        assert scene.shade.shape[0] == nb_vertices
        assert scene.colors.shape[0] == nb_vertices
        assert scene.colors.shape[1] == nb_colors
        assert scene.edgeflags.shape[0] == nb_triangles
        assert scene.edgeflags.shape[1] == 3
        assert scene.textured.shape[0] == nb_triangles
        assert scene.shaded.shape[0] == nb_triangles

        assert (scene.background_image is not None) != (
            scene.background_color is not None
        )

        if scene.background_image is not None:
            assert scene.background_image.ndim == 3
            assert scene.background_image.shape[0] == height
            assert scene.background_image.shape[1] == width
            assert scene.background_image.shape[2] == nb_colors
            assert (
                scene.background_color is None
            ), "You need to provide either background_image or background_color"
        else:
            assert (
                scene.background_color is not None
            ), "You need to provide background_image or background_color"
            assert scene.background_color.shape[0] == nb_colors

        if scene.texture.size > 0:
            assert scene.texture.ndim == 3
            assert scene.texture.shape[0] > 0
            assert scene.texture.shape[1] > 0
            assert scene.texture.shape[2] == nb_colors

        assert z_buffer.shape[0] == height
        assert z_buffer.shape[1] == width

        if antialiase_error:
            assert err_buffer is not None, "You need to provide err_buffer"
            assert obs is not None, "You need to provide obs"
            assert err_buffer.shape[0] == height
            assert err_buffer.shape[1] == width
            assert obs.shape[0] == height
            assert obs.shape[1] == width
            assert obs.shape[2] == nb_colors

    differentiable_renderer_cython.renderSceneCpp(
        scene, sigma, image, z_buffer, antialiase_error, obs, err_buffer
    )


def renderSceneB(
    scene: Scene2DBase,
    sigma: float,
    image: np.ndarray,
    z_buffer: np.ndarray,
    image_b: Optional[np.ndarray] = None,
    antialiase_error: bool = False,
    obs: Optional[np.ndarray] = None,
    err_buffer: Optional[np.ndarray] = None,
    err_buffer_b: Optional[np.ndarray] = None,
    check_valid: bool = True,
) -> None:
    if check_valid:
        # doing checks here as it seems the debugger in not able to find the pyx file
        # when installed from a wheel. this also make interactive debugging easier
        # for the library user

        assert image is not None
        assert z_buffer is not None

        height = image.shape[0]
        width = image.shape[1]
        nb_colors = image.shape[2]
        nb_triangles = scene.faces.shape[0]

        assert nb_colors == scene.colors.shape[1]
        assert z_buffer.shape[0] == height
        assert z_buffer.shape[1] == width
        assert nb_triangles == scene.faces_uv.shape[0]

        nb_vertices = scene.depths.shape[0]
        nb_vertices_uv = scene.uv.shape[0]

        assert scene.faces.dtype == np.uint32
        assert np.all(scene.faces < nb_vertices)
        assert np.all(scene.faces_uv < nb_vertices_uv)

        assert scene.colors.ndim == 2
        assert scene.uv.ndim == 2
        assert scene.ij.ndim == 2
        assert scene.shade.ndim == 1
        assert scene.edgeflags.ndim == 2
        assert scene.textured.ndim == 1
        assert scene.shaded.ndim == 1
        assert scene.uv.shape[1] == 2
        assert scene.ij.shape[0] == nb_vertices
        assert scene.ij.shape[1] == 2
        assert scene.shade.shape[0] == nb_vertices
        assert scene.colors.shape[0] == nb_vertices
        assert scene.colors.shape[1] == nb_colors
        assert scene.edgeflags.shape[0] == nb_triangles
        assert scene.edgeflags.shape[1] == 3
        assert scene.textured.shape[0] == nb_triangles
        assert scene.shaded.shape[0] == nb_triangles

        assert (scene.background_image is not None) != (
            scene.background_color is not None
        )

        if scene.background_image is not None:
            assert scene.background_image.ndim == 3
            assert scene.background_image.shape[0] == height
            assert scene.background_image.shape[1] == width
            assert scene.background_image.shape[2] == nb_colors
            assert (
                scene.background_color is None
            ), "You need to provide either background_image or background_color"
        else:
            assert (
                scene.background_color is not None
            ), "You need to provide background_image or background_color"
            assert scene.background_color.shape[0] == nb_colors

        assert scene.uv_b is not None
        assert scene.ij_b is not None
        assert scene.shade_b is not None
        assert scene.uv_b is not None
        assert scene.colors_b is not None

        assert scene.uv_b.ndim == 2
        assert scene.ij_b.ndim == 2
        assert scene.shade_b.ndim == 1
        assert scene.edgeflags.ndim == 2
        assert scene.textured.ndim == 1
        assert scene.shaded.ndim == 1
        assert scene.uv_b.shape[0] == nb_vertices_uv
        assert scene.uv_b.shape[1] == 2
        assert scene.ij_b.shape[0] == nb_vertices
        assert scene.ij_b.shape[1] == 2
        assert scene.shade_b.shape[0] == nb_vertices
        assert scene.colors_b.shape[0] == nb_vertices
        assert scene.colors_b.shape[1] == nb_colors

        if scene.texture.size > 0:
            assert scene.texture_b is not None
            assert scene.texture.ndim == 3
            assert scene.texture_b.ndim == 3
            assert scene.texture.shape[0] > 0
            assert scene.texture.shape[1] > 0
            assert scene.texture.shape[0] == scene.texture_b.shape[0]
            assert scene.texture.shape[1] == scene.texture_b.shape[1]
            assert scene.texture.shape[2] == nb_colors
            assert scene.texture_b.shape[2] == nb_colors

        if antialiase_error:
            assert err_buffer is not None
            assert obs is not None
            assert err_buffer.shape[0] == height
            assert err_buffer.shape[1] == width
            assert obs.shape[0] == height
            assert obs.shape[1] == width
        else:
            assert image_b is not None
            assert image_b.shape[0] == height
            assert image_b.shape[1] == width

    differentiable_renderer_cython.renderSceneBCpp(
        scene,
        sigma,
        image,
        z_buffer,
        image_b,
        antialiase_error,
        obs,
        err_buffer,
        err_buffer_b,
    )


class Camera:
    """Camera class with the same distortion parameterization as OpenCV."""

    def __init__(
        self,
        extrinsic: np.ndarray,
        intrinsic: np.ndarray,
        height: int,
        width: int,
        distortion: Union[Optional[np.ndarray], Iterable[float]] = None,
        checks: bool = True,
        tol: float = 1e-6,
    ):
        if checks:
            assert extrinsic.shape == (3, 4)
            assert intrinsic.shape == (3, 3)
            assert np.all(intrinsic[2, :] == [0, 0, 1])
            assert (
                np.linalg.norm(extrinsic[:3, :3].T.dot(extrinsic[:3, :3]) - np.eye(3))
                < tol
            )

            if distortion is not None:
                distortion = np.array(distortion)
                assert distortion.shape == (5,)

        self.extrinsic = extrinsic
        self.intrinsic = intrinsic
        self.distortion = distortion
        self.height = height
        self.width = width

    @property
    def xfov(self) -> float:
        assert self.intrinsic[0, 2] == self.width / 2
        return np.degrees(2 * np.arctan(self.width / (2 * self.intrinsic[0, 0])))

    @property
    def yfov(self) -> float:
        assert self.intrinsic[1, 2] == self.height / 2
        return np.degrees(2 * np.arctan(self.height / (2 * self.intrinsic[1, 1])))

    def world_to_camera(self, points_3d: np.ndarray) -> np.ndarray:
        assert points_3d.shape[-1] == 3
        return points_3d.dot(self.extrinsic[:3, :3].T) + self.extrinsic[:3, 3]

    def camera_to_world_mtx_4x4(self) -> np.ndarray:
        return np.row_stack(
            (
                np.column_stack((self.extrinsic[:, :3].T, self.get_center())),
                np.array((0, 0, 0, 1)),
            )
        )

    def left_mul_intrinsic(self, projected: np.ndarray) -> np.ndarray:
        assert projected.ndim == 2
        assert projected.shape[-1] == 2
        return projected.dot(self.intrinsic[:2, :2].T) + self.intrinsic[:2, 2]

    def column_stack(
        self, values: Union[List[np.ndarray], Tuple[np.ndarray, ...]]
    ) -> np.ndarray:
        return np.column_stack(values)

    @overload
    def project_points(
        self,
        points_3d: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def project_points(
        self,
        points_3d: np.ndarray,
        *,
        store_backward: Optional[Dict[str, Any]] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def project_points(
        self,
        points_3d: np.ndarray,
        return_depths: Literal[True],
        store_backward: Optional[Dict[str, Any]] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def project_points(
        self,
        points_3d: np.ndarray,
        return_depths: Literal[False],
        store_backward: Optional[Dict[str, Any]] = None,
    ) -> np.ndarray:
        ...

    def project_points(
        self,
        points_3d: np.ndarray,
        return_depths: bool = True,
        store_backward: Optional[Dict[str, Any]] = None,
    ) -> Union[
        Tuple[np.ndarray, np.ndarray], np.ndarray
    ]:  # similar to cv2.project_points
        p_camera = self.world_to_camera(points_3d)
        depths = p_camera[:, 2]
        projected = p_camera[:, :2] / depths[:, None]

        if self.distortion is None:
            projected_image_coordinates = self.left_mul_intrinsic(projected)
            if store_backward is not None:
                store_backward["project_points"] = (p_camera, depths, projected)
        else:
            (
                k1,
                k2,
                p1,
                p2,
                k3,
            ) = self.distortion
            x = projected[:, 0]
            y = projected[:, 1]
            x2 = x**2
            y2 = y**2
            r2 = x2 + y2
            r4 = r2 * r2
            r6 = r2 * r4
            radial_distortion = 1 + k1 * r2 + k2 * r4 + k3 * r6
            tangential_distortion_x = 2 * p1 * x * y + p2 * (r2 + 2 * x2)
            tangential_distortion_y = p1 * (r2 + 2 * y2) + 2 * p2 * x * y
            distorted_x = x * radial_distortion + tangential_distortion_x
            distorted_y = y * radial_distortion + tangential_distortion_y
            distorted = self.column_stack((distorted_x, distorted_y))
            projected_image_coordinates = self.left_mul_intrinsic(distorted)
            if store_backward is not None:
                store_backward["project_points"] = (
                    p_camera,
                    depths,
                    projected,
                    r2,
                    radial_distortion,
                )

        if return_depths:
            return projected_image_coordinates, depths
        else:
            return projected_image_coordinates

    def project_points_backward(
        self,
        projected_image_coordinates_b: np.ndarray,
        store_backward: Dict[str, Any],
        depths_b: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        if self.distortion is None:
            p_camera, depths, projected = store_backward["project_points"]
            projected_b = projected_image_coordinates_b.dot(
                self.intrinsic[:2, :2].T
            )  # not sure about transpose

        else:
            p_camera, depths, projected, r2, radial_distortion = store_backward[
                "project_points"
            ]
            (
                k1,
                k2,
                p1,
                p2,
                k3,
            ) = self.distortion
            x = projected[:, 0]
            y = projected[:, 1]
            distorted_b = projected_image_coordinates_b.dot(
                self.intrinsic[:2, :2].T
            )  # not sure about transpose
            distorted_x_b = distorted_b[:, 0]
            distorted_y_b = distorted_b[:, 1]
            x_b = distorted_x_b * radial_distortion
            y_b = distorted_y_b * radial_distortion
            radial_distortion_b = distorted_x_b * x + distorted_y_b * y
            tangential_distortion_x_b = distorted_x_b
            tangential_distortion_y_b = distorted_y_b
            x_b += tangential_distortion_x_b * (2 * p1 * y + p2 * 4 * x)
            y_b += tangential_distortion_x_b * 2 * p1 * x
            x_b += tangential_distortion_y_b * 2 * p2 * y
            y_b += tangential_distortion_y_b * (2 * p2 * x + p1 * 4 * y)
            r2_b = tangential_distortion_x_b * p2 + tangential_distortion_y_b * p1
            r2_b += radial_distortion_b * (k1 + 2 * k2 * r2 + 3 * k3 * r2**2)
            x_b += r2_b * 2 * x
            y_b += r2_b * 2 * y
            projected_b = np.column_stack((x_b, y_b))

        p_camera_b = np.column_stack(
            (
                projected_b / depths[:, None],
                -np.sum(projected_b * p_camera[:, :2], axis=1) / (depths**2),
            )
        )
        if depths_b is not None:
            p_camera_b[:, 2] += depths_b
        return p_camera_b.dot(self.extrinsic[:3, :3].T)

    def get_center(self) -> np.ndarray:
        return -self.extrinsic[:3, :3].T.dot(self.extrinsic[:, 3])

    def __repr__(self) -> str:
        return (
            f"<Camera>\n"
            f"width:\n{str(self.width)}\n"
            f"height:\n{str(self.height)}\n"
            f"extrinsic:\n{str(self.extrinsic)}\n"
            f"intrinsic:\n{str(self.intrinsic)}\n"
            f"distortion:\n{str(self.distortion)}\n"
        )


class PerspectiveCamera(Camera):
    """Camera with perspective projection."""

    def __init__(
        self,
        width: int,
        height: int,
        fov: float,
        camera_center: np.ndarray,
        rot: Optional[np.ndarray] = None,
        distortion: Optional[np.ndarray] = None,
    ):
        """Perspective camera constructor.

        - width: width of the camera in pixels
        - height: eight of the camera in pixels
        - fov: horizontal field of view in degrees
        - camera_center: center of the camera in world coordinate system
        - rot: 3x3 rotation matrix word to camera (x_cam = rot.dot(x_world))\
            default to identity
        - distortion: distortion parameters
        """
        assert camera_center.shape == (3,)
        if rot is None:
            rot = np.eye(3)
        else:
            # check it is indeed a rotation matrix
            assert rot.shape == (3, 3)
            assert np.allclose(rot.T.dot(rot), np.eye(3), 1e-6)
            assert np.linalg.det(rot) > 0
        focal = 0.5 * width / np.tan(0.5 * fov * np.pi / 180)
        focal_x = focal
        pixel_aspect_ratio = 1
        focal_y = focal * pixel_aspect_ratio
        trans = -rot.T.dot(camera_center)
        cx = width / 2
        cy = height / 2
        intrinsic = np.array([[focal_x, 0, cx], [0, focal_y, cy], [0, 0, 1]])
        extrinsic = np.column_stack((rot, trans))
        super().__init__(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            distortion=distortion,
            width=width,
            height=height,
        )


def default_camera(
    width: int,
    height: int,
    fov: float,
    vertices: np.ndarray,
    rot: np.ndarray,
    distortion: Optional[np.ndarray] = None,
) -> Camera:
    """Compute the position of the camera center so that the entire mesh is visible
    and covers most or the image.
    """
    cam_vertices = vertices.dot(rot.T)
    box_min = cam_vertices.min(axis=0)
    box_max = cam_vertices.max(axis=0)
    box_center = 0.5 * (box_max + box_min)
    box_size = box_max - box_min
    camera_distance_x = (
        0.5 * box_size[0] / np.tan(0.5 * fov * np.pi / 180) + 0.5 * box_size[2]
    )
    camera_distance_y = (
        0.5 * box_size[1] * (width / height) / np.tan(0.5 * fov * np.pi / 180)
        + 0.5 * box_size[2]
    )
    camera_distance = max(camera_distance_x, camera_distance_y)
    camera_center = rot.T.dot(box_center + np.array([0, 0, -camera_distance]))
    return PerspectiveCamera(width, height, fov, camera_center, rot, distortion)


class Scene2D(Scene2DBase):
    """Class representing a 2.5D scene. It contains a set of 2D vertices with
    associated depths and a list of faces that are triplets of vertices indexes.
    """

    def __init__(
        self,
        faces: np.ndarray,
        faces_uv: np.ndarray,
        ij: np.ndarray,
        depths: np.ndarray,
        textured: np.ndarray,
        uv: np.ndarray,
        shade: np.ndarray,
        colors: np.ndarray,
        shaded: np.ndarray,
        edgeflags: np.ndarray,
        height: int,
        width: int,
        nb_colors: int,
        texture: np.ndarray,
        background_image: Optional[np.ndarray] = None,
        background_color: Optional[np.ndarray] = None,
        clockwise: bool = False,
        backface_culling: bool = False,
        strict_edge: bool = True,
        perspective_correct: bool = False,
        integer_pixel_centers: bool = True,
    ):
        """
        Conventions:
        If integer_pixel_centers is True (default) then pixel centers are at integer coordinates with
            upper left at (0, 0)
            upper right at (width - 1, 0)
            lower left at (0, height - 1)
            lower right at  (width - 1, height - 1)
        if integer_pixel_centers is False, then pixel centers are at half integer coordinates with
            upper left at (0.5, 0.5)
            upper right at (width - 0.5, 0.5)
            lower left at (0.5, height - 0.5)
            lower right at  (width -0.5, height - 0.5)
        """

        self.faces = faces
        self.faces_uv = faces_uv
        self.ij = ij
        self.depths = depths
        self.textured = textured
        self.uv = uv
        self.shade = shade
        self.colors = colors
        self.shaded = shaded
        self.edgeflags = edgeflags
        self.height = height
        self.width = width
        self.nb_colors = nb_colors
        self.texture = texture
        self.background_image = background_image
        self.background_color = background_color
        self.clockwise = clockwise
        self.backface_culling = backface_culling
        self.strict_edge = strict_edge
        self.perspective_correct = perspective_correct
        self.integer_pixel_centers = integer_pixel_centers

        # fields to store gradients
        self.uv_b = np.zeros(self.uv.shape)
        self.ij_b = np.zeros(self.ij.shape)
        self.shade_b = np.zeros(self.shade.shape)
        self.colors_b = np.zeros(self.colors.shape)
        self.texture_b = np.zeros(self.texture.shape)
        self.store_backward: Tuple

    def clear_gradients(self) -> None:
        assert self.uv_b is not None
        assert self.ij_b is not None
        assert self.shade_b is not None
        assert self.colors_b is not None
        assert self.texture_b is not None

        self.uv_b.fill(0)
        self.ij_b.fill(0)
        self.shade_b.fill(0)
        self.colors_b.fill(0)
        self.texture_b.fill(0)

    def render_error(
        self, obs: np.ndarray, sigma: float = 1
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        image = np.zeros((self.height, self.width, self.nb_colors))
        z_buffer = np.zeros((self.height, self.width))
        err_buffer = np.empty((self.height, self.width))
        antialiase_error = True
        renderScene(self, sigma, image, z_buffer, antialiase_error, obs, err_buffer)
        self.store_backward = (sigma, obs, image, z_buffer, err_buffer)
        return image, z_buffer, err_buffer

    def render(self, sigma: float = 1) -> Tuple[np.ndarray, np.ndarray]:
        image = np.zeros((self.height, self.width, self.nb_colors))
        z_buffer = np.zeros((self.height, self.width))
        antialiase_error = False
        renderScene(self, sigma, image, z_buffer, antialiase_error, None, None)
        self.store_backward = (sigma, image, z_buffer)
        return image, z_buffer

    def render_error_backward(
        self, err_buffer_b: np.ndarray, make_copies: bool = True
    ) -> None:
        if self.perspective_correct:
            raise BaseException(
                "perspective_correct not supported yet for gradient back propagation"
            )

        if not self.backface_culling:
            raise BaseException(
                "use backface_culling=True if you use gradient backpropagation to get valid gradient through edge anti-aliasing."
            )
        sigma, obs, image, z_buffer, err_buffer = self.store_backward
        antialiase_error = True
        if make_copies:
            renderSceneB(
                self,
                sigma,
                image,
                z_buffer,
                None,
                antialiase_error,
                obs,
                err_buffer.copy(),
                err_buffer_b,
            )
        else:
            renderSceneB(
                self,
                sigma,
                image,
                z_buffer,
                None,
                antialiase_error,
                obs,
                err_buffer,
                err_buffer_b,
            )

    def render_backward(self, image_b: np.ndarray, make_copies: bool = True) -> None:
        if self.perspective_correct:
            raise BaseException(
                "perspective_correct not supported yet for gradient back propagation"
            )
        if not self.backface_culling:
            raise BaseException(
                "use backface_culling=True if you use gradient backpropagation to get valid gradient through edge anti-aliasing."
            )
        sigma, image, z_buffer = self.store_backward
        antialiase_error = False
        if (
            make_copies
        ):  # if we make copies we keep the antialized image unchanged image
            # along the occlusion boundaries
            renderSceneB(
                self,
                sigma,
                image.copy(),
                z_buffer,
                image_b,
                antialiase_error,
                None,
                None,
                None,
            )
        else:
            renderSceneB(
                self,
                sigma,
                image,
                z_buffer,
                image_b,
                antialiase_error,
                None,
                None,
                None,
            )

    def render_compare_and_backward(
        self,
        obs: np.ndarray,
        sigma: float = 1,
        antialiase_error: bool = False,
        mask: Optional[np.ndarray] = None,
        clear_gradients: bool = True,
        make_copies: bool = True,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, float]:
        if self.perspective_correct:
            raise BaseException(
                "perspective_correct not supported yet for gradient back propagation"
            )
        if mask is None:
            mask = np.ones((obs.shape[0], obs.shape[1]))
        if antialiase_error:
            image, z_buffer, err_buffer = self.render_error(obs, sigma)
        else:
            image, z_buffer = self.render(sigma)

        if clear_gradients:
            self.clear_gradients()

        if antialiase_error:
            err_buffer = err_buffer * mask
            err = float(np.sum(err_buffer))
            err_buffer_b = copy.copy(mask)
            self.render_error_backward(err_buffer_b, make_copies=make_copies)
        else:
            diff_image = (image - obs) * mask[:, :, None]
            err_buffer = (diff_image) ** 2
            err = float(np.sum(err_buffer))
            image_b = 2 * diff_image
            self.render_backward(image_b, make_copies=make_copies)

        return image, z_buffer, err_buffer, err


class Scene3D:
    """Class representing a 3D scene containing a single mesh, a directional light
    and an ambient light. The parameter sigma control the width of
    antialiasing edge overdraw.
    """

    def __init__(
        self,
        sigma: float = 1,
        perspective_correct: bool = False,
        integer_pixel_centers: bool = True,
    ):
        self.mesh: Optional[ColoredTriMesh] = None

        self.light_directional: Optional[np.ndarray] = None
        self.light_ambient: float = 0
        self.sigma = sigma
        self.perspective_correct = perspective_correct
        self.background_image: Optional[np.ndarray] = None
        self.background_color: Optional[np.ndarray] = None
        self.integer_pixel_centers = integer_pixel_centers
        self.colors: Optional[np.ndarray] = None

        self.store_backward_current: Optional[Dict[str, Any]] = None
        self.vertices_b: Optional[np.ndarray] = None
        self.scene2D: Optional[Scene2D] = None

    def clear_gradients(self) -> None:
        # fields to store gradients
        assert self.mesh is not None
        self.scene_2d.clear_gradients()

    def set_light(
        self,
        light_directional: Union[Tuple[float, float, float], np.ndarray],
        light_ambient: float,
    ) -> None:
        """
        light_ambient : scalar. Intensity of the ambient light
        light_directional : 3d vector. Directional light are at an infinite distance and thus
        there is no position.  The light_directional vector corresponds to the  direction
        multiplied by the intensity (instead of a normalized direction and a scalar intensity).
        This  parameterization has been chosen because it makes it easier to do gradient
        descent  as there is not normalization constraint. However it does not support colored lights.
        """
        if light_directional is not None:
            self.light_directional = np.array(light_directional)
        else:
            self.light_directional = None
        self.light_ambient = light_ambient

    def set_mesh(self, mesh: ColoredTriMesh) -> None:
        self.mesh = mesh

    def set_background(self, background_image: np.ndarray) -> None:
        warnings.warn(
            "This will be deprecated, please use set_background_image or set_background_color",
            UserWarning,
        )
        self.set_background_image(background_image)

    def set_background_image(self, background_image: np.ndarray) -> None:
        if self.background_color is not None:
            raise BaseException(
                "you cannot provide both background image and background color"
            )
        background_image = np.asanyarray(background_image)
        assert background_image.dtype == np.double
        assert background_image.ndim == 3
        self.background_image = background_image

    def set_background_color(
        self, background_color: Union[Iterable[float], np.ndarray]
    ) -> None:
        if self.background_image is not None:
            raise BaseException(
                "you cannot provide both background image and background color"
            )
        background_color = np.asanyarray(background_color, dtype=np.float64)
        assert background_color.dtype == np.double
        assert background_color.ndim == 1
        self.background_color = background_color

    def compute_vertices_luminosity(self) -> np.ndarray:
        assert self.mesh is not None
        if self.light_directional is not None:
            directional = np.maximum(
                0, -np.sum(self.mesh.vertex_normals * self.light_directional, axis=1)
            )
        else:
            directional = np.zeros((self.mesh.nb_vertices))
        if self.store_backward_current is not None:
            self.store_backward_current["compute_vertices_luminosity"] = directional
        return directional + self.light_ambient

    def _compute_vertices_colors_with_illumination(self) -> np.ndarray:
        assert self.mesh is not None
        vertices_luminosity = self.compute_vertices_luminosity()
        if self.store_backward_current is not None:
            self.store_backward_current[
                "_compute_vertices_colors_with_illumination"
            ] = vertices_luminosity
        return self.mesh.vertices_colors * vertices_luminosity[:, None]

    def _compute_vertices_colors_with_illumination_backward(
        self, colors_b: np.ndarray
    ) -> None:
        assert self.mesh is not None
        assert self.store_backward_current is not None
        vertices_luminosity: np.ndarray = self.store_backward_current[
            "_compute_vertices_colors_with_illumination"
        ]
        vertices_luminosity_b = np.sum(self.mesh.vertices_colors * colors_b, axis=1)
        self.mesh.vertices_colors_b = colors_b * vertices_luminosity[:, None]

        self.compute_vertices_luminosity_backward(vertices_luminosity_b)

    def compute_vertices_luminosity_backward(
        self, vertices_luminosity_b: np.ndarray
    ) -> None:
        assert self.mesh is not None
        assert self.store_backward_current is not None
        if self.light_directional is not None:
            directional: np.ndarray = self.store_backward_current[
                "compute_vertices_luminosity"
            ]
            self.light_directional_b = -np.sum(
                ((vertices_luminosity_b * (directional > 0))[:, None])
                * self.mesh.vertex_normals,
                axis=0,
            )
            self.vertex_normals_b = (
                -((vertices_luminosity_b * (directional > 0))[:, None])
                * self.light_directional
            )
        self.light_ambient_b = np.sum(vertices_luminosity_b)

    def _render_2d(self) -> Tuple[np.ndarray, np.ndarray]:
        nb_color_channels = self.scene_2d.nb_colors
        image = np.empty((self.scene_2d.height, self.scene_2d.width, nb_color_channels))
        z_buffer = np.empty((self.scene_2d.height, self.scene_2d.width))

        renderScene(self.scene_2d, self.sigma, image, z_buffer)

        if self.store_backward_current is not None:
            self.store_backward_current["render_2d"] = (image, z_buffer)

        return image, z_buffer

    def _render_2d_backward(self, image_b: np.ndarray) -> None:
        if self.perspective_correct:
            raise BaseException(
                "perspective_correct not supported yet for gradient back propagation"
            )
        assert self.store_backward_current is not None
        image, z_buffer = self.store_backward_current["render_2d"]
        # self.ij = np.array(ij)
        # self.colors = np.array(colors)
        renderSceneB(self.scene_2d, self.sigma, image.copy(), z_buffer, image_b)
        self.scene_2D = None

    @overload
    def render(
        self,
        camera: Camera,
    ) -> np.ndarray:
        ...

    @overload
    def render(
        self,
        camera: Camera,
        return_z_buffer: Literal[True],
        backface_culling: bool = True,
    ) -> Tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def render(
        self,
        camera: Camera,
        return_z_buffer: Literal[False],
        backface_culling: bool = True,
    ) -> np.ndarray:
        ...

    def render(
        self,
        camera: Camera,
        return_z_buffer: bool = False,
        backface_culling: bool = True,
    ) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
        assert self.mesh is not None, "You need to provide a mesh first."
        self.store_backward_current = {}

        if self.light_directional is not None:
            self.mesh.compute_vertex_normals()

        if (self.background_image is None) == (self.background_color is None):
            raise BaseException(
                "You need to provide either a background image or background color"
            )

        points_2d, depths = camera.project_points(
            self.mesh.vertices, store_backward=self.store_backward_current
        )

        # compute silhouette edges
        if self.sigma > 0:
            edgeflags = self.mesh.edge_on_silhouette(points_2d)
        else:
            edgeflags = np.zeros((self.mesh.nb_faces, 3), dtype=bool)
        # construct 2D scene
        faces = self.mesh.faces.astype(np.uint32)

        if self.mesh.uv is not None:
            assert self.mesh.texture is not None
            uv = self.mesh.uv
            faces_uv = self.mesh.faces_uv
            textured = np.ones((self.mesh.nb_faces), dtype=bool)
            shade = self.compute_vertices_luminosity()
            shaded = np.ones(
                (self.mesh.nb_faces), dtype=bool
            )  # could eventually be non zero if we were using texture
            texture = self.mesh.texture
            nb_colors = texture.shape[2]
            colors = np.zeros((self.mesh.nb_vertices, nb_colors))
        else:
            colors = self._compute_vertices_colors_with_illumination()
            nb_colors = colors.shape[1]
            faces_uv = faces
            uv = np.zeros((self.mesh.nb_vertices, 2))
            textured = np.zeros((self.mesh.nb_faces), dtype=bool)
            shade = np.zeros(
                (self.mesh.nb_vertices), dtype=np.float64
            )  # could eventually be non zero if we were using texture
            shaded = np.zeros(
                (self.mesh.nb_faces), dtype=bool
            )  # could eventually be non zero if we were using texture
            texture = np.zeros((0, 0))

        height = camera.height
        width = camera.width
        strict_edge = True

        clockwise = self.mesh.clockwise

        assert faces_uv is not None  # helping mypy

        scene_2d = Scene2D(
            faces=faces,
            faces_uv=faces_uv,
            ij=points_2d,
            depths=depths,
            textured=textured,
            uv=uv,
            shade=shade,
            colors=colors,
            shaded=shaded,
            edgeflags=edgeflags,
            height=height,
            width=width,
            nb_colors=nb_colors,
            texture=texture,
            background_image=self.background_image,
            background_color=self.background_color,
            clockwise=clockwise,
            backface_culling=backface_culling,
            strict_edge=strict_edge,
            perspective_correct=self.perspective_correct,
            integer_pixel_centers=self.integer_pixel_centers,
        )
        self.scene_2d = scene_2d

        image, z_buffer = self._render_2d()
        if self.store_backward_current is not None:
            self.store_backward_current["render"] = (
                camera,
                edgeflags,
            )  # store this field as it could be overwritten when
            # rendering several views
        return (image, z_buffer) if return_z_buffer else image

    def render_backward(self, image_b: np.ndarray) -> None:
        assert self.scene_2d is not None
        assert self.scene_2d.colors_b is not None
        assert self.mesh is not None
        if self.perspective_correct:
            raise BaseException(
                "perspective_correct not supported yet for gradient back propagation"
            )
        assert self.store_backward_current is not None
        camera, self.edgeflags = self.store_backward_current["render"]
        self._render_2d_backward(image_b)
        self._compute_vertices_colors_with_illumination_backward(self.scene_2d.colors_b)
        self.mesh._vertices_b = camera.project_points_backward(
            self.scene_2d.ij_b, store_backward=self.store_backward_current
        )
        if self.light_directional is not None:
            self.mesh.compute_vertex_normals_backward(self.vertex_normals_b)

    def render_depth(
        self, camera: Camera, depth_scale: float = 1, backface_culling: bool = True
    ) -> np.ndarray:
        assert self.mesh is not None
        self.store_backward_current = {}
        points_2d, depths = camera.project_points(
            self.mesh.vertices, store_backward=self.store_backward_current
        )

        # compute silhouette edges
        if self.sigma > 0:
            edgeflags = self.mesh.edge_on_silhouette(points_2d)
        else:
            edgeflags = np.zeros((self.mesh.nb_faces, 3), dtype=bool)

        scene_2d = Scene2D(
            ij=points_2d,
            nb_colors=1,
            faces=self.mesh.faces.astype(np.uint32),
            faces_uv=self.mesh.faces,
            colors=depths[:, None] * depth_scale,
            depths=depths,
            edgeflags=edgeflags,
            uv=np.zeros((self.mesh.nb_vertices, 2)),
            textured=np.zeros((self.mesh.nb_faces), dtype=bool),
            shade=np.zeros(
                (self.mesh.nb_vertices), dtype=bool
            ),  # eventually used when using texture
            height=camera.height,
            width=camera.width,
            shaded=np.zeros(
                (self.mesh.nb_faces), dtype=bool
            ),  # eventually used when using texture
            texture=np.zeros((0, 0)),
            clockwise=self.mesh.clockwise,
            backface_culling=backface_culling,
            strict_edge=True,
            background_color=self.background_color,
            integer_pixel_centers=self.integer_pixel_centers,
        )
        self.scene_2d = scene_2d
        image, _ = self._render_2d()
        if self.store_backward_current is not None:
            self.store_backward_current["render_depth"] = (camera, depth_scale)
        return image

    def render_depth_backward(self, depth_b: np.ndarray) -> None:
        assert self.store_backward_current is not None
        assert self.mesh is not None
        if self.perspective_correct:
            raise BaseException(
                "perspective_correct not supported yet for gradient back propagation"
            )
        camera, depth_scale = self.store_backward_current["render_depth"]
        self._render_2d_backward(depth_b)
        depths_b = np.squeeze(self.scene_2d.colors_b * depth_scale, axis=1)
        self.mesh._vertices_b = camera.project_points_backward(
            self.scene_2d.ij_b,
            depths_b=depths_b,
            store_backward=self.store_backward_current,
        )

    def render_deferred(
        self,
        camera: Camera,
        depth_scale: float = 1,
        color: bool = True,
        depth: bool = True,
        face_id: bool = True,
        barycentric: bool = True,
        normal: bool = True,
        luminosity: bool = True,
        uv: bool = True,
        xyz: bool = True,
        backface_culling: bool = True,
    ) -> Dict[str, np.ndarray]:
        assert self.mesh is not None, "You need to provide a mesh first"
        points_2d, depths = camera.project_points(self.mesh.vertices)

        # compute silhouette edges
        self.store_backward_current = None

        if self.sigma > 0:
            raise BaseException(
                "Antialiasing is not supposed to be used when using deferred rendering, please use sigma==0"
            )

        edgeflags = np.zeros((self.mesh.nb_faces, 3), dtype=bool)

        if luminosity or normal:
            self.mesh.compute_vertex_normals()
        if luminosity:
            vertices_luminosity = self.compute_vertices_luminosity()

        # construct triangle soup (loosing connectivity), needed to render
        # discontinuous uv maps and face ids
        soup_nb_faces = self.mesh.nb_faces
        soup_nb_vertices = 3 * self.mesh.nb_faces
        soup_faces = np.arange(0, soup_nb_vertices, dtype=np.uint32).reshape(
            self.mesh.nb_faces, 3
        )

        soup_faces_uv = soup_faces
        soup_ij = points_2d[self.mesh.faces].reshape(soup_nb_vertices, 2)
        soup_depths = depths[self.mesh.faces].reshape(soup_nb_vertices, 1)

        channels = {}
        if depth:
            channels["depth"] = soup_depths * depth_scale
        if face_id:
            soup_face_ids = np.tile(
                np.arange(0, self.mesh.nb_faces)[:, None], (1, 3)
            ).reshape(soup_nb_vertices, 1)
            channels["face_id"] = soup_face_ids
        if barycentric:
            soup_barycentric = np.tile(
                np.eye(3, 3)[None, :, :], (self.mesh.nb_faces, 1, 1)
            ).reshape(soup_nb_vertices, 3)
            channels["barycentric"] = soup_barycentric
        if normal:
            soup_normals = self.mesh.vertex_normals[self.mesh.faces].reshape(
                soup_nb_vertices, 3
            )
            channels["normal"] = soup_normals
        if luminosity:
            soup_luminosity = vertices_luminosity[self.mesh.faces].reshape(
                soup_nb_vertices, 1
            )
            channels["luminosity"] = soup_luminosity
        if xyz:
            soup_xyz = self.mesh.vertices[self.mesh.faces].reshape(soup_nb_vertices, 3)
            channels["xyz"] = soup_xyz

        if self.mesh.uv is None:
            if color:
                assert self.mesh.vertices_colors is not None
                soup_vertices_colors = self.mesh.vertices_colors[
                    self.mesh.faces
                ].reshape(soup_nb_vertices, 3)
                channels["color"] = soup_vertices_colors
        elif uv:
            soup_uv = self.mesh.uv[self.mesh.faces_uv].reshape(soup_nb_vertices, 2)
            channels["uv"] = soup_uv

        offset = 0
        ranges = {}
        for k, v in channels.items():
            size = v.shape[1]
            ranges[k] = (offset, offset + size)
            offset += size

        colors = np.column_stack(list(channels.values()))

        nb_colors = colors.shape[1]
        uv_zeros = np.zeros((soup_nb_vertices, 2))
        textured = np.zeros((soup_nb_faces), dtype=bool)
        shade = np.zeros((soup_nb_vertices), dtype=bool)

        height = camera.height
        width = camera.width
        shaded = np.zeros(
            (soup_nb_faces), dtype=bool
        )  # eventually used when using texture
        texture = np.zeros((0, 0))

        background_image = np.zeros((height, width, nb_colors))
        if "depth" in channels:
            background_image[
                :, :, ranges["depth"][0] : ranges["depth"][1]
            ] = depths.max()

        scene_2d = Scene2DBase(
            faces=soup_faces,
            faces_uv=soup_faces_uv,
            ij=soup_ij,
            depths=soup_depths,
            textured=textured,
            uv=uv_zeros,
            shade=shade,
            colors=colors,
            shaded=shaded,
            edgeflags=edgeflags,
            height=height,
            width=width,
            nb_colors=nb_colors,
            texture=texture,
            background_image=background_image,
            backface_culling=backface_culling,
            background_color=None,
            integer_pixel_centers=self.integer_pixel_centers,
        )
        buffers = np.empty((camera.height, camera.width, nb_colors))
        z_buffer = np.empty((camera.height, camera.width))
        renderScene(scene_2d, 0, buffers, z_buffer)

        output: Dict[str, np.ndarray] = {
            k: buffers[:, :, ranges[k][0] : ranges[k][1]] for k in channels
        }

        return output



"""Tensorflow implementation of an as-rigid-as-possible energy based on the difference of laplacian with a reference shape."""
from typing import Tuple

import numpy as np
import tensorflow as tf

from ..laplacian_rigid_energy import LaplacianRigidEnergy
from .tools import scipy_sparse_matrix_to_tensorflow
from .triangulated_mesh_tensorflow import ColoredTriMeshTensorflow


class LaplacianRigidEnergyTensorflow:
    """Tensorflow class that implements an as-rigid-as-possible energy based on the difference of laplacian with a reference shape."""

    def __init__(
        self, mesh: ColoredTriMeshTensorflow, vertices: np.ndarray, cregu: float
    ):
        self.numpy_imp = LaplacianRigidEnergy(mesh, vertices, cregu)
        self.cT_tf = scipy_sparse_matrix_to_tensorflow(self.numpy_imp.cT)
        self.approx_hessian = scipy_sparse_matrix_to_tensorflow(
            self.numpy_imp.approx_hessian
        )

    def evaluate(
        self, vertices: tf.Tensor
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.SparseTensor]:
        assert isinstance(vertices, tf.Tensor)
        diff = tf.reshape(vertices - tf.constant(self.numpy_imp.vertices_ref), [-1])
        grad_vertices = tf.reshape(
            self.numpy_imp.cregu
            * tf.sparse.sparse_dense_matmul(self.cT_tf, diff[:, None]),
            vertices.shape,
        )  # 5 times slower than scipy
        energy = 0.5 * tf.reduce_sum(diff * tf.reshape(grad_vertices, [-1]))

        return energy, grad_vertices, self.approx_hessian



"""OpenGL sharders used to render scenes with camera distortion."""

vertex_shader_source = """
#version 140
uniform mat4 intrinsic;
uniform mat4 extrinsic;
uniform float k1;
uniform float k2;
uniform float k3;
uniform float p1;
uniform float p2;

in vec3 in_vert;
in vec3 in_norm;
in vec3 in_text;

out vec3 v_vert;
out vec3 v_norm;
out vec3 v_text;

void main() {
        v_vert = in_vert;
        v_norm = in_norm;
        v_text = in_text;
        vec4 p_camera = extrinsic* vec4(v_vert, 1.0);
        vec2 projected = p_camera.xy/p_camera.z;
        float x = projected.x;
        float y = projected.y;
        float r2 = x * x + y * y;
        float radial_distortion = 1.0 + k1 * r2 + k2 * r2 * r2 + k3 * r2 * r2 *r2;
        float tangential_distortion_x = 2.0 * p1 * x * y + p2 * (r2 + 2.0 * x * x);
        float tangential_distortion_y = p1 * (r2 + 2.0 * y * y) + 2.0 * p2 * x * y;
        float distorted_x = x * radial_distortion + tangential_distortion_x;
        float distorted_y = y * radial_distortion + tangential_distortion_y;
        p_camera.xy = vec2(distorted_x, distorted_y)*p_camera.z;
        gl_Position=intrinsic*p_camera;
}
"""

fragment_shader_rgb_source = """
#version 130
uniform sampler2D Texture;
uniform vec4 Color;
uniform vec3 light_directional;
uniform float light_ambient;
in vec3 v_vert;
in vec3 v_norm;
in vec3 v_text;

out vec4 f_color;

void main() {
        float lum = light_ambient + max(dot(normalize(v_norm),- light_directional),0.0);
        vec3 color = texture(Texture, v_text.xy).rgb;
        color = color * (1.0 - Color.a) + Color.rgb * Color.a;
        f_color = vec4(color * lum, 1.0);
}"""



import os
from typing import List
from xml.dom import minidom

from imageio.v3 import imwrite
import numpy as np

from .differentiable_renderer import Camera
from .obj import save_obj
from .triangulated_mesh import ColoredTriMesh


def export_meshlab(
    filename: str,
    mesh: ColoredTriMesh,
    cameras: List[Camera],
    images: List[np.ndarray],
    obj_name: str = "mesh.obj",
) -> None:
    root = minidom.Document()
    xml = root.createElement("MeshLabProject")
    root.appendChild(xml)

    mesh_group = root.createElement("MeshGroup")
    xml.appendChild(mesh_group)
    ml_mesh = root.createElement("MLMesh")
    ml_mesh.setAttribute("filename", obj_name)
    ml_mesh.setAttribute("visible", "1")
    ml_mesh.setAttribute("label", obj_name)
    save_obj(
        os.path.join(os.path.dirname(filename), obj_name), mesh.vertices, mesh.faces
    )

    mesh_group.appendChild(ml_mesh)
    ml_matrix = root.createElement("MLMatrix")
    txt = root.createTextNode("1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1")
    ml_matrix.appendChild(txt)
    mesh_group.appendChild(ml_matrix)

    render_group = root.createElement("RasterGroup")

    for i, (camera, image) in enumerate(zip(cameras, images)):
        image_file = f"raster{i:d}.png"
        ml_raster = root.createElement("MLRaster")
        vcg_camera = root.createElement("VCGCamera")
        mtx = camera.camera_to_world_mtx_4x4()
        translation = np.diag([-1, -1, -1, 1]).dot(mtx[:, 3])
        vcg_camera.setAttribute(
            "TranslationVector",
            " ".join([str(v) for v in translation]),
        )

        vcg_camera.setAttribute(
            "CenterPx", " ".join([str(v) for v in camera.intrinsic[0:2, 2]])
        )
        vcg_camera.setAttribute("PixelSizeMm", "1 1")
        vcg_camera.setAttribute("FocalMm", str(camera.intrinsic[0, 0]))
        vcg_camera.setAttribute("LensDistortion", "0 0")
        vcg_camera.setAttribute("CameraType", "0")
        vcg_camera.setAttribute("BinaryData", "0")
        vcg_camera.setAttribute("ViewportPx", f"{image.shape[0]} {image.shape[1]}")
        rotation = np.diag([1, -1, -1, 1]).dot(mtx)
        rotation[:3, 3] = 0
        vcg_camera.setAttribute(
            "RotationMatrix", " ".join([str(v) for v in rotation.flatten()])
        )

        ml_raster.appendChild(vcg_camera)
        plane = root.createElement("Plane")

        plane.setAttribute("fileName", image_file)
        plane.setAttribute("semantic", "1")
        ml_raster.appendChild(plane)

        imwrite(image_file, image.astype(np.uint8))
        ml_mesh.setAttribute("label", image_file)
        render_group.appendChild(ml_raster)

    xml.appendChild(render_group)
    # <VCGCamera TranslationVector="0.136953 0.683761 -2.68297 1" CenterPx="698 592" PixelSizeMm="0.0369161 0.0369161" FocalMm="61.0115" LensDistortion="0 0" CameraType="0" BinaryData="0" ViewportPx="1397 1184" RotationMatrix="1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1 "/>
    # <Plane fileName="hand.png" semantic="1"/>
    # </MLRaster>

    xml_str = root.toprettyxml(indent="\t")

    with open(filename, "w") as f:
        f.write(xml_str)
    # <!DOCTYPE MeshLabDocument>
    # <MeshLabProject>
    # <MeshGroup>
    # <MLMesh filename="hand.obj" visible="1" label="hand.obj">
    # <MLMatrix44>
    # 1 0 0 0
    # 0 1 0 0
    # 0 0 1 0
    # 0 0 0 1
    # </MLMatrix44>
    # <RenderingOption wireWidth="1" wireColor="64 64 64 255" pointSize="3" pointColor="131 149 69 255" solidColor="192 192 192 255" boxColor="234 234 234 255">100001000000000000000100000001010100000010100000000100111011110000001001</RenderingOption>
    # </MLMesh>
    # </MeshGroup>
    # <RasterGroup>
    # <MLRaster label="hand.png">
    # <VCGCamera TranslationVector="0.136953 0.683761 -2.68297 1" CenterPx="698 592" PixelSizeMm="0.0369161 0.0369161" FocalMm="61.0115" LensDistortion="0 0" CameraType="0" BinaryData="0" ViewportPx="1397 1184" RotationMatrix="1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1 "/>
    # <Plane fileName="hand.png" semantic="1"/>
    # </MLRaster>

    # <MLRaster label="hand.png">
    # <VCGCamera TranslationVector="0.136953 0.683761 -4.68297 1" CenterPx="698 592" PixelSizeMm="0.0369161 0.0369161" FocalMm="61.0115" LensDistortion="0 0" CameraType="0" BinaryData="0" ViewportPx="1397 1184" RotationMatrix="1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1 "/>
    # <Plane fileName="hand.png" semantic="1"/>
    # </MLRaster>
    # </RasterGroup>
    # </MeshLabProject>



"""Tooling functions for 3D geometry processing with their backward derivatives implementations."""

from typing import Callable, Tuple

import numpy as np


def qrot(q: np.ndarray, v: np.ndarray) -> np.ndarray:
    assert q.ndim in [1, 2]
    assert q.shape[-1] == 4
    assert v.ndim == 2
    assert v.shape[-1] == 3

    if q.ndim == 2:
        uv = np.cross(q[:, None, :3], v[None, :, :])
        uuv = np.cross(q[:, None, :3], uv)
        return v + 2 * (q[:, None, [3]] * uv + uuv)
    else:
        uv = np.cross(q[:3], v)
        uuv = np.cross(q[:3], uv)
        return v + 2 * (q[3] * uv + uuv)


def qrot_backward(
    q: np.ndarray, v: np.ndarray, vr_b: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    uv = np.cross(q[:3], v)
    v_b = vr_b.copy()
    q_b = np.zeros((4))
    q_b[3] = 2 * np.sum(vr_b * uv)
    uuv_b = 2 * vr_b.copy()
    uv_b = 2 * vr_b * q[3] + np.cross(uuv_b, q[:3])
    q_b[:3] = np.sum(np.cross(uv, uuv_b), axis=0) + np.sum(np.cross(v, uv_b), axis=0)
    v_b += np.cross(uv_b, q[:3])
    return q_b, v_b


def normalize(x: np.ndarray, axis: int = -1) -> np.ndarray:
    n2 = np.sum(x**2, axis=axis)
    n = np.sqrt(n2)
    return x / np.expand_dims(n, axis)


def normalize_backward(x: np.ndarray, xn_b: np.ndarray, axis: int = -1) -> np.ndarray:
    n2 = np.sum(x**2, axis=axis)
    n = np.sqrt(n2)
    inv_n = 1 / n
    n_b = -np.sum(xn_b * x, axis=axis) * (inv_n**2)
    return (xn_b + x * np.expand_dims(n_b, axis)) * np.expand_dims(inv_n, axis)


def cross_backward(
    u: np.ndarray, v: np.ndarray, c_b: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    v_b = np.cross(c_b, u)
    u_b = np.cross(v, c_b)
    return u_b, v_b


def jacobian_finite_differences(
    func: Callable[[np.ndarray], np.ndarray], x: np.ndarray, epsilon: float = 1e-6
) -> np.ndarray:
    v0 = func(x)
    nx = x.copy()
    jacobian = np.zeros((v0.size, x.size))
    nx_flat = nx.ravel()
    for d in range(x.size):
        v = nx_flat[d]
        nx_flat[d] = v + epsilon
        d1 = func(nx)
        nx_flat[d] = v - epsilon
        d2 = func(nx)
        nx_flat[d] = v
        jacobian[:, d] = (d1 - d2).flatten() / (2 * epsilon)
    v02 = func(x)
    assert np.allclose(
        v0, v02
    ), "The value of the function changed. The function has a state or is non deterministic"
    return jacobian


def check_jacobian_finite_differences(
    jac: np.ndarray,
    func: Callable[[np.ndarray], np.ndarray],
    x: np.ndarray,
    epsilon: float = 1e-7,
    tol: float = 1e-4,
) -> None:
    nx = x.copy()
    nx_flat = nx.ravel()
    for d in range(x.size):
        nx_flat[d] = x.flat[d] + epsilon
        d1 = func(nx)
        nx_flat[d] = x.flat[d] - epsilon
        d2 = func(nx)
        nx_flat[d] = x.flat[d]
        jac_cold_fd = (d1 - d2).flatten() / (2 * epsilon)
        max_diff = np.max(np.abs(jac[..., d] - jac_cold_fd))
        assert max_diff < tol



"""Example with fitting a 32 triangles soup to an image."""

import copy
import hashlib
import os
from typing import Any, Dict, List, Tuple

import cv2
from imageio.v3 import imread
import matplotlib.pyplot as plt
import numpy as np

import deodr
from deodr import differentiable_renderer_cython  # type: ignore
from deodr.differentiable_renderer import Scene2D


def create_example_scene(
    n_tri: int = 30,
    width: int = 200,
    height: int = 200,
    clockwise: bool = False,
    textured_ratio: float = 0.5,
) -> Scene2D:
    material = (
        imread(os.path.join(deodr.data_path, "trefle.jpg")).astype(np.float64) / 255
    )
    height_material = material.shape[0]
    width_material = material.shape[1]

    scale_matrix = np.array([[height, 0], [0, width]])
    scale_material = np.array([[height_material - 1, 0], [0, width_material - 1]])

    triangles = []
    for _ in range(n_tri):
        tmp = scale_matrix.dot(
            np.random.rand(2, 1).dot(np.ones((1, 3)))
            + 0.5 * (-0.5 + np.random.rand(2, 3))
        )
        while np.abs(np.linalg.det(np.vstack((tmp, np.ones((3)))))) < 1500:
            tmp = scale_matrix.dot(
                np.random.rand(2, 1).dot(np.ones((1, 3)))
                + 0.5 * (-0.5 + np.random.rand(2, 3))
            )

        if np.linalg.det(np.vstack((tmp, np.ones((3))))) > 0:
            tmp = np.fliplr(tmp)
        triangle = {
            "ij": tmp.T,
            "depths": (np.random.rand(1) * np.ones((3, 1))),
        }

        triangle["textured"] = np.random.rand(1) > (1 - textured_ratio)

        if triangle["textured"]:
            triangle["uv"] = (
                scale_material.dot(np.array([[0, 1, 0.2], [0, 0.2, 1]])).T + 1
            )  # texture coordinate of the vertices
            triangle["shade"] = np.random.rand(3, 1)  # shade  intensity at each vertex
            triangle["colors"] = np.zeros((3, 3))
            triangle["shaded"] = True
        else:
            triangle["uv"] = np.zeros((3, 2))
            triangle["shade"] = np.zeros((3, 1))
            triangle["colors"] = np.random.rand(3, 3)
            # colors of the vertices (can be gray, rgb color,or even other dimension
            # vectors) when using simple linear interpolation across triangles
            triangle["shaded"] = False

        triangle["edgeflags"] = np.array(
            [True, True, True]
        )  # all edges are discontinuity edges as no triangle pair share an edge
        triangles.append(triangle)

    scene: Dict[str, Any] = {
        key: np.squeeze(np.vstack([np.array(triangle[key]) for triangle in triangles]))
        for key in [
            "ij",
            "depths",
            "textured",
            "uv",
            "shade",
            "colors",
            "shaded",
            "edgeflags",
        ]
    }

    scene["faces"] = np.arange(3 * n_tri).reshape(-1, 3).astype(np.uint32)
    scene["faces_uv"] = np.arange(3 * n_tri).reshape(-1, 3).astype(np.uint32)

    if clockwise:
        scene["faces"] = np.fliplr(scene["faces"])
        scene["faces_uv"] = np.fliplr(scene["faces_uv"])

    scene["clockwise"] = clockwise
    scene["height"] = height
    scene["width"] = width
    scene["texture"] = material
    scene["nb_colors"] = 3
    scene["background_color"] = None
    scene["background_image"] = np.tile(
        np.array([0.3, 0.5, 0.7])[None, None, :], (height, width, 1)
    )
    scene["perspective_correct"] = False
    scene["backface_culling"] = True
    return Scene2D(**scene)


def run(
    nb_max_iter: int = 500,
    display: bool = True,
    clockwise: bool = False,
    antialiase_error: bool = False,
) -> Tuple[List[float], List[str]]:
    print("process id=%d" % os.getpid())

    np.random.seed(2)
    scene_gt = create_example_scene(clockwise=clockwise)
    sigma = 1

    image_target = np.zeros((scene_gt.height, scene_gt.width, scene_gt.nb_colors))

    z_buffer = np.zeros((scene_gt.height, scene_gt.width))
    differentiable_renderer_cython.renderSceneCpp(
        scene_gt, sigma, image_target, z_buffer
    )

    n_vertices = len(scene_gt.depths)

    displacement_magnitude_ij = 10
    displacement_magnitude_uv = 0
    displacement_magnitude_colors = 0

    alpha_ij = 0.01
    beta_ij = 0.80
    alpha_uv = 0.03
    beta_uv = 0.80
    alpha_color = 0.001
    beta_color = 0.70

    max_uv = np.array(scene_gt.texture.shape[:2]) - 1

    scene_init = copy.deepcopy(scene_gt)
    scene_init.ij = (
        scene_gt.ij + np.random.randn(n_vertices, 2) * displacement_magnitude_ij
    )
    scene_init.uv = (
        scene_gt.uv + np.random.randn(n_vertices, 2) * displacement_magnitude_uv
    )
    scene_init.uv = np.maximum(scene_init.uv, 0)
    scene_init.uv = np.minimum(scene_init.uv, max_uv)
    scene_init.colors = (
        scene_gt.colors + np.random.randn(n_vertices, 3) * displacement_magnitude_colors
    )

    hashes: List[str] = []

    np.random.seed(2)
    scene_iter = copy.deepcopy(scene_init)

    speed_ij = np.zeros((n_vertices, 2))
    speed_uv = np.zeros((n_vertices, 2))
    speed_color = np.zeros((n_vertices, 3))

    losses: List[float] = []
    for niter in range(nb_max_iter):
        image, depth, loss_image, loss = scene_iter.render_compare_and_backward(
            sigma=sigma, antialiase_error=antialiase_error, obs=image_target
        )
        hashes.append(hashlib.sha256(image.tobytes()).hexdigest())
        print(f"iter {niter} loss = {loss}")
        # imwrite(os.path.join(iterfolder,f'soup_{niter}.png'), combinedIMage)

        losses.append(loss)
        if loss_image.ndim == 2:
            loss_image = np.broadcast_to(loss_image[:, :, None], image.shape)
        if display:
            cv2.waitKey(1)
            cv2.imshow(
                "animation",
                np.column_stack((image_target, image, loss_image))[:, :, ::-1],
            )

        assert scene_iter.colors_b is not None
        assert scene_iter.ij_b is not None
        assert scene_iter.uv_b is not None
        if displacement_magnitude_ij > 0:
            speed_ij = beta_ij * speed_ij - scene_iter.ij_b * alpha_ij
            scene_iter.ij = scene_iter.ij + speed_ij

        if displacement_magnitude_colors > 0:
            speed_color = beta_color * speed_color - scene_iter.colors_b * alpha_color
            scene_iter.colors = scene_iter.colors + speed_color

        if displacement_magnitude_uv > 0:
            speed_uv = beta_uv * speed_uv - scene_iter.uv_b * alpha_uv
            scene_iter.uv = scene_iter.uv + speed_uv
            scene_iter.uv = np.maximum(scene_iter.uv, 0.0)
            scene_iter.uv = np.maximum(scene_iter.uv, max_uv)

    return losses, hashes


def run_with_and_without_antialiased_error(display: bool = True) -> None:
    if display:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    for antialiase_error in [False, True]:
        if display:
            losses, _ = run(
                nb_max_iter=500,
                display=True,
                clockwise=False,
                antialiase_error=antialiase_error,
            )
            ax.plot(losses, label="antialiaseError=%d" % antialiase_error)
    if display:
        plt.legend()
        plt.show()


if __name__ == "__main__":
    run_with_and_without_antialiased_error()



"""Test texture coordinates regarding 0.5 texel shift."""


import numpy as np

from deodr.differentiable_renderer import Scene2D


def test_texture_coordinates() -> None:
    texture = np.array(
        [[[1, 0, 0], [0, 1, 0]], [[0, 0, 1], [1, 1, 1]]], dtype=np.float64
    )

    # coordinate of upper left pixel center is (0,0)
    # coordinate of the upper left texture pixel center (texel) is (0, 0)
    # the color of the texture bilinearly sampled as (0,0) is texture[0, 0]
    # this contrasts with OpenGL when the texture at position (0.5/texture_width , 0.5/texture_width) is texture[0, 0]

    uv = np.array([[0, 0], [1, 0], [0, 1]])
    ij = np.array([[1, 1], [1, 15], [15, 1]])
    depths = np.array([1, 1, 1])
    shade = np.array([0, 0, 0])
    shade = np.array([1, 1, 1])
    textured = np.array([1], dtype=bool)
    shaded = np.array([1], dtype=bool)
    colors = np.eye(3)
    edgeflags = np.zeros((1, 3), dtype=bool)

    for clockwise in [False, True]:
        if clockwise:
            faces = np.array([[0, 2, 1]], dtype=np.uint32)
            faces_uv = np.array([[0, 2, 1]], dtype=np.uint32)
        else:
            faces = np.array([[0, 1, 2]], dtype=np.uint32)
            faces_uv = np.array([[0, 1, 2]], dtype=np.uint32)

        height = 40
        width = 60

        background_color = np.array([0, 0, 0])

        # using strict_edge == True so that the pixel that are exactly on the edge
        # are drawn
        scene2D = Scene2D(
            ij=ij,
            faces=faces,
            faces_uv=faces_uv,
            uv=uv,
            texture=texture,
            height=height,
            width=width,
            nb_colors=3,
            background_image=None,
            background_color=background_color,
            depths=depths,
            textured=textured,
            shade=shade,
            colors=colors,
            shaded=shaded,
            edgeflags=edgeflags,
            strict_edge=False,
            perspective_correct=True,
            clockwise=clockwise,
        )

        image, _ = scene2D.render(sigma=0)
        assert np.allclose(image[0, :, :], [0, 0, 0])
        assert np.allclose(image[:, 0, :], [0, 0, 0])
        assert np.allclose(image[1, 1, :], [1, 0, 0])
        assert np.allclose(image[15, 1, :], [0, 1, 0])
        assert np.allclose(image[1, 15, :], [0, 0, 1])


if __name__ == "__main__":
    test_texture_coordinates()



"""example fitting a texture mesh to an image using face images"""

from typing import Dict, Tuple

import cv2
import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial import Delaunay
from sklearn import decomposition
import sklearn.datasets

from deodr.differentiable_renderer import Scene2D


def main() -> None:
    faces = sklearn.datasets.fetch_olivetti_faces()

    # plt.imshow(faces.images[0])

    faces_pca = decomposition.PCA(n_components=150, whiten=True)
    faces_pca.fit(faces.data)
    plt.imshow(faces_pca.mean_.reshape(faces.images[0].shape), cmap=plt.cm.bone)

    # coefs = faces_pca.transform(faces.data)
    # print(faces)

    # faces_pca.inverse_transform(coefs[[0]])

    # create a regular grid and a triangulation  of the 2D image
    N = 5
    points = np.column_stack(
        [t.flatten() for t in np.meshgrid(np.arange(N + 1) / N, np.arange(N + 1) / N)]
    )
    tri = Delaunay(points)
    triangles = tri.simplices.astype(np.uint32)
    plt.triplot(points[:, 0], points[:, 1], triangles)
    on_border = np.any((points == 0) | (points == 1), axis=1)

    # deform the grid
    max_displacement = 0.5
    np.random.seed(0)
    points_deformed_gt = (
        points + (np.random.rand(*points.shape) - 0.5) * max_displacement / N
    )
    points_deformed_gt[on_border] = points[on_border]
    plt.triplot(points_deformed_gt[:, 0], points_deformed_gt[:, 1], triangles)
    # plt.show()
    nb_points = points.shape[0]
    nb_triangles = triangles.shape[0]
    ij = points_deformed_gt * 64 - 0.5
    uv = points * 64 + 0.5
    textured = np.ones((nb_triangles), dtype=bool)
    shaded = np.ones((nb_triangles), dtype=bool)
    depths = np.ones((nb_points))
    shade = np.ones((nb_points))
    colors = np.ones((nb_points, 1))
    edgeflags = np.zeros((nb_triangles, 3), dtype=bool)
    height = 64
    width = 64
    nb_colors = 1
    texture = faces.images[10][:, :, None]
    background_color = np.array([0])

    scene_gt = Scene2D(
        triangles,
        triangles,
        ij,
        depths,
        textured,
        uv,
        shade,
        colors,
        shaded,
        edgeflags,
        height,
        width,
        nb_colors,
        texture,
        background_image=None,
        background_color=background_color,
        clockwise=True,
        backface_culling=True,
    )

    image_gt, _ = scene_gt.render(sigma=1)
    plt.subplot(3, 1, 1)
    plt.imshow(np.squeeze(texture, axis=2))
    plt.subplot(3, 1, 2)
    plt.imshow(np.squeeze(image_gt, axis=2))
    plt.subplot(3, 1, 3)
    plt.imshow(np.squeeze(np.abs(image_gt - texture), axis=2))

    # plt.show()
    np.max(texture - image_gt)

    scene = Scene2D(
        triangles,
        triangles,
        ij,
        depths,
        textured,
        uv,
        shade,
        colors,
        shaded,
        edgeflags,
        height,
        width,
        nb_colors,
        texture,
        background_image=None,
        background_color=background_color,
        clockwise=True,
        backface_culling=True,
    )
    # cv2.namedWindow('animation',cv2.WINDOW_NORMAL)
    # cv2.resizeWindow('animation', 600,600)

    rescale_factor = 10

    def fun(
        points_deformed: np.ndarray, pca_coefs: np.ndarray
    ) -> Tuple[float, Dict[str, np.ndarray]]:
        ij = points_deformed * 64 - 0.5
        # face=faces_pca.inverse_transform(coefs[:,None])
        face = (faces_pca.mean_ + pca_coefs.dot(faces_pca.components_)).reshape(
            (64, 64)
        )
        scene.ij = ij
        scene.texture = face[:, :, None]
        print("render")
        image, z_buffer, diff_image, err = scene.render_compare_and_backward(
            obs=image_gt, sigma=1
        )
        assert scene.ij_b is not None  # helping mypy
        print("np.max(np.abs(scene.ij_b))=%f" % np.max(np.abs(scene.ij_b)))
        print("E=%f" % err)
        print("done")
        image_gt_zoomed = cv2.resize(
            (image_gt.copy() * 255).astype(np.uint8),
            None,
            fx=rescale_factor,
            fy=rescale_factor,
            interpolation=cv2.INTER_NEAREST,
        )
        image_zoomed = cv2.resize(
            (image.copy() * 255).astype(np.uint8),
            None,
            fx=rescale_factor,
            fy=rescale_factor,
            interpolation=cv2.INTER_NEAREST,
        )
        diff_image_zoomed = cv2.resize(
            (np.abs(diff_image) * 255 * 10).astype(np.uint8),
            None,
            fx=rescale_factor,
            fy=rescale_factor,
            interpolation=cv2.INTER_NEAREST,
        )

        cv2.polylines(
            image_gt_zoomed,
            (points_deformed_gt[tri.simplices] * 64 * rescale_factor).astype(np.int32),
            isClosed=True,
            color=(0, 0, 0),
            lineType=cv2.LINE_AA,
        )

        cv2.polylines(
            image_zoomed,
            (points_deformed[tri.simplices] * 64 * rescale_factor).astype(np.int32),
            isClosed=True,
            color=(0, 0, 0),
            lineType=cv2.LINE_AA,
        )
        cv2.imshow(
            "animation",
            np.column_stack((image_gt_zoomed, image_zoomed, diff_image_zoomed)),
        )
        cv2.waitKey(1)

        # get gradient on pca coefs
        assert scene.texture_b is not None
        assert scene.ij_b is not None
        coefs_grad = faces_pca.components_.dot(scene.texture_b.flatten())
        points_deformed_grad = scene.ij_b * 64
        print(np.max(np.abs(points_deformed_grad)))
        grads = {"points_deformed": points_deformed_grad, "pca_coefs": coefs_grad}
        return err, grads

    nb_iter = 100

    pca_coefs = np.zeros((faces_pca.n_components))
    points_deformed = points.copy()

    variables = {"points_deformed": points_deformed, "pca_coefs": pca_coefs}
    lambdas = {"points_deformed": 0.0001, "pca_coefs": 0.5}

    for niter in range(nb_iter):
        E, grads = fun(**variables)
        print(f"iter{niter} E={E}")
        for name in variables:
            variables[name] = variables[name] - lambdas[name] * grads[name]

        variables["points_deformed"][on_border] = points[on_border]

    cv2.waitKey()


if __name__ == "__main__":
    main()



"""Test using depth image hand fitting."""

import numpy as np
import tensorflow as tf

from deodr.examples.depth_image_hand_fitting import run


def test_depth_image_hand_fitting_pytorch() -> None:
    energies = run(
        dl_library="pytorch",
        plot_curves=False,
        display=False,
        save_images=False,
        max_iter=50,
    )

    possible_results = [
        251.32711067513003,
        251.31652686512888,
        251.31652686495823,
    ]

    assert np.any(np.abs(np.array(possible_results) - energies[49]) < 1e-5)


def test_depth_image_hand_fitting_numpy() -> None:
    energies = run(
        dl_library="none",
        plot_curves=False,
        display=False,
        save_images=False,
        max_iter=50,
    )

    possible_results = [
        251.32711113732933,
        251.32711113730954,
        251.3271111242092,
    ]

    assert np.any(np.abs(np.array(possible_results) - energies[49]) < 1e-5)


def test_depth_image_hand_fitting_tensorflow() -> None:
    tf.config.set_visible_devices(
        [], "GPU"
    )  # Running on CPU to get determinisic results

    energies = run(
        dl_library="tensorflow",
        plot_curves=False,
        display=False,
        save_images=False,
        max_iter=50,
    )

    possible_results = [
        251.31648932312913,
        251.3164914350016,
        251.3164891265543,
        251.32711038915872,
    ]

    assert np.any(np.abs(np.array(possible_results) - energies[49]) < 1e-5)


if __name__ == "__main__":
    test_depth_image_hand_fitting_pytorch()
    test_depth_image_hand_fitting_numpy()

    test_depth_image_hand_fitting_tensorflow()



"""Implementation of an as-rigid-as-possible energy based on the difference of laplacian with a reference shape."""
import copy
from typing import Tuple

import numpy as np
import scipy
import scipy.sparse

from deodr.triangulated_mesh import TriMesh


class LaplacianRigidEnergy:
    """Class that implements an as-rigid-as-possible energy based on the difference of laplacian with a reference shape."""

    def __init__(self, mesh: TriMesh, vertices: np.ndarray, cregu: float):
        self.cT = scipy.sparse.kron(
            mesh.adjacencies.laplacian.T * mesh.adjacencies.laplacian,
            scipy.sparse.eye(3),
        ).tocsr()
        self.vertices_ref = copy.copy(vertices)
        self.mesh = mesh
        self.cregu = cregu
        self.approx_hessian = self.cregu * self.cT
        n_components, _ = scipy.sparse.csgraph.connected_components(
            csgraph=self.mesh.adjacencies.adjacency_vertices,
            directed=False,
            return_labels=True,
        )
        if n_components > 1:
            raise (
                BaseException(
                    "You have more than one connected component in your mesh."
                )
            )

    def evaluate(
        self,
        vertices: np.ndarray,
    ) -> Tuple[float, np.ndarray, scipy.sparse.csr_matrix]:
        diff = (vertices - self.vertices_ref).flatten()
        grad_vertices = self.cregu * (self.cT * diff).reshape((vertices.shape[0], 3))
        energy = 0.5 * diff.dot(grad_vertices.flatten())

        return energy, grad_vertices, self.approx_hessian



"""Test using triangle soup rendering."""
from typing import Any, Dict, List

from deodr.examples.triangle_soup_fitting import run


def check_results(
    all_lkg_results: List[Dict[str, Any]],
    losses: List[float],
    hashes: List[str],
    check_hashes: bool,
) -> None:
    for lkg_results in all_lkg_results:
        valid = all(
            losses[key] == value for key, value in lkg_results["losses"].items()
        )

        if check_hashes:
            for key, value in lkg_results["hashes"].items():
                if hashes[key] != value:
                    valid = False
        if valid:
            return
    raise BaseException("None of the results matches")


def test_triangle_soup_fitting(check_hashes: bool = True) -> None:
    losses, hashes = run(
        nb_max_iter=50, display=False, clockwise=False, antialiase_error=False
    )

    lkg_results = [
        {  # windows
            "losses": {-1: 1331.3578738815468},
            "hashes": {
                0: "38b6f6954374230aeb1ce5d804308522f6b4c58a6736a040aeef7f2176a20b28",
                1: "0434ea722edb9e3364da9b0e8564c3002b9aa3b12791ba8f089689beecd3c4e9",
            },
        },
        {  # google colab Intel(R) Xeon(R) CPU @ 2.20GHz: 251.3164914350016
            "losses": {-1: 1328.235645237829},
            "hashes": {
                0: "f0571dcd10df3c81c902b703462b3c94139eab3d88bc8c081c1dca1c2533fc4c",
                1: "fe8416554d92528a13ebed3cb0d9d376abfbce6171c18d831ed765b0bbc7198a",
            },
        },
    ]

    check_results(lkg_results, losses, hashes, check_hashes)

    losses, hashes = run(
        nb_max_iter=50, display=False, clockwise=False, antialiase_error=True
    )

    lkg_results = [
        {
            # windows
            "losses": {-1: 1457.8585914203582},
            "hashes": {
                0: "82a7b73fde3615ef7c70008965f4bfda8610b9001c20dd435a880bf45a31d3d6",
                1: "0de2e8b80730cfc444d0552cd81e5071897a525ec6495e643ca17fb0792496c0",
            },
        },
        {  # google colab Intel(R) Xeon(R) CPU @ 2.20GHz: 251.3164914350016
            "losses": {-1: 1456.4362294737318},
            "hashes": {
                0: "55b545a8ae70b51ed0ad762210daf7b3bf568c786f4ddd5716ca6ec6479951ba",
                1: "e8c598184d89fe966ffe3efdfc2ca2500191f92b1e40d71f9213847c90982a1f",
            },
        },
    ]

    check_results(lkg_results, losses, hashes, check_hashes)

    losses, hashes = run(
        nb_max_iter=50, display=False, clockwise=True, antialiase_error=False
    )

    lkg_results = [
        {  # windows
            "losses": {-1: 1331.357873881545},
            "hashes": {
                0: "eb9f335acdeb2abc3e029826a56fd77bd7a4bb39c2794d9972735a9f388ddeba",
                1: "6b4cc11e56dfebe2b8235485ea1bb8230a5765e7f0eeb2ac671d9cd4e3311f74",
            },
        },
        {  # google colab Intel(R) Xeon(R) CPU @ 2.20GHz: 251.3164914350016
            "losses": {-1: 1328.2356452378326},
            "hashes": {
                0: "427c434a1e60537c57aa2d6a5a990e425c064fa9690955014f2dd7c087b30cfc",
                1: "6882ed3113f6b33505deeb5366d75ec48ffece0870af9ab73df74e9aa5c8f870",
            },
        },
    ]

    check_results(lkg_results, losses, hashes, check_hashes)
    losses, hashes = run(
        nb_max_iter=50, display=False, clockwise=True, antialiase_error=True
    )

    lkg_results = [
        {
            "losses": {-1: 1457.8585914203607},
            "hashes": {
                0: "d2113e271f29afbdac393767297b923caaa4fe19af4bdd16d3b649c2f5ba3103",
                1: "8820a5f90b4c0496b88542f42ed773afcd7cc0e0e3edc7a3da6427e2ab881d44",
            },
        },
        {  # google colab Intel(R) Xeon(R) CPU @ 2.20GHz: 251.3164914350016
            "losses": {-1: 1456.4362294737307},
            "hashes": {
                0: "0b41270c59ddf37d8fd7e9ff0f0723042bd6ee0546ac071dadb306f65aa143ad",
                1: "11e225715108f8fa4a307bd5c8ed2d1891abaec83290a8fcb3677e58d258ccfd",
            },
        },
    ]

    check_results(lkg_results, losses, hashes, check_hashes)


if __name__ == "__main__":
    test_triangle_soup_fitting()



# -*- coding: utf-8 -*-
"""deodr

Discontinuity-Edge-Overdraw Differentiable Rendering

If you use this library please cite
Model-based 3D Hand Pose Estimation from Monocular Video.
M. de la Gorce, N. Paragios and David Fleet. PAMI 2011

Martin de La Gorce. 2019.
"""
__version__ = "0.2.1"

__all__ = [
    "Scene2D",
    "Scene3D",
    "Camera",
    "read_obj",
    "LaplacianRigidEnergy",
    "TriMesh",
    "ColoredTriMesh",
]

import os

from .differentiable_renderer import Camera, Scene2D, Scene3D
from .laplacian_rigid_energy import LaplacianRigidEnergy
from .obj import read_obj
from .triangulated_mesh import ColoredTriMesh, TriMesh

root_path = os.path.dirname(__file__)

data_path = os.path.join(root_path, "data")



"""Setup script for the DEODR project."""

import os
import re

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup

extension = Extension(
    name="deodr.differentiable_renderer_cython",
    sources=["deodr/differentiable_renderer_cython.pyx"],
    include_dirs=["deodr", np.get_include()],
)

with open(os.path.join(os.path.dirname(__file__), "deodr", "__init__.py")) as fp:
    for line in fp:
        m = re.search(r'^\s*__version__\s*=\s*([\'"])([^\'"]+)\1\s*$', line)
        if m:
            version = m.group(2)
            break
    else:
        raise RuntimeError("Unable to find own __version__ string")
print(f"version = {version}")

setup(
    version=version,
    url="https://github.com/martinResearch/DEODR",
    data_files=[("C++", ["C++/DifferentiableRenderer.h"])],
    ext_modules=cythonize([extension], annotate=True, build_dir="build"),
    include_dirs=[np.get_include()],
    packages=find_packages(),
    package_data={"deodr": ["*.pyx", "*.pxd", "data/*.*", "data/**/*.*"]},
    long_description=open("readme.md", "r").read(),
    long_description_content_type="text/markdown",
)



"""Modules containing Tensorflow classes to fit 3D meshes to images using differentiable rendering."""
import copy
from typing import Optional, Tuple, Union

import numpy as np
import scipy.sparse.linalg
import scipy.spatial.transform.rotation
import tensorflow as tf

from . import CameraTensorflow, LaplacianRigidEnergyTensorflow, Scene3DTensorflow
from .. import LaplacianRigidEnergy
from .triangulated_mesh_tensorflow import ColoredTriMeshTensorflow as ColoredTriMesh


def qrot(q: tf.Tensor, v: tf.Tensor) -> tf.Tensor:
    qr = tf.tile(q[None, :], (v.shape[0], 1))
    qvec = qr[:, :-1]
    uv = tf.linalg.cross(qvec, v)
    uuv = tf.linalg.cross(qvec, uv)
    return v + 2 * (qr[:, 3][None, 0] * uv + uuv)


def mult_and_clamp(x: Union[np.ndarray, tf.Tensor], a: float, t: float) -> np.ndarray:
    return np.minimum(np.maximum(x * a, -t), t)


class MeshDepthFitter:
    """Class to fit a deformable mesh to a depth image using Tensorflow."""

    def __init__(
        self,
        vertices: np.ndarray,
        faces: np.ndarray,
        euler_init: np.ndarray,
        translation_init: np.ndarray,
        cregu: float = 2000,
        inertia: float = 0.96,
        damping: float = 0.05,
    ):
        self.cregu = cregu
        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 0.5
        self.step_factor_quaternion = 0.00006
        self.step_max_quaternion = 0.1
        self.step_factor_translation = 0.00005
        self.step_max_translation = 0.1

        self.mesh = ColoredTriMesh(
            faces, vertices, colors=np.zeros((vertices.shape[0], 0))
        )  # we do a copy to avoid negative stride not support by Tensorflow
        object_center = vertices.mean(axis=0)
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([-0.5, 0, 5]) * object_radius

        self.scene = Scene3DTensorflow()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergy(self.mesh, vertices, cregu)
        self.vertices_init = tf.constant(copy.copy(vertices))
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)
        self.reset()

    def set_mesh_transform_init(
        self, euler: np.ndarray, translation: np.ndarray
    ) -> None:
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self) -> None:
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices_init.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

    def set_max_depth(self, max_depth: float) -> None:
        self.max_depth = max_depth
        self.scene.set_background_color([max_depth])

    def set_depth_scale(self, depth_scale: float) -> None:
        self.depthScale = depth_scale

    def set_image(
        self,
        mesh_image: np.ndarray,
        focal: Optional[float] = None,
        distortion: Optional[np.ndarray] = None,
    ) -> None:
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 2
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = CameraTensorflow(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            width=self.width,
            height=self.height,
            distortion=distortion,
        )
        self.iter = 0

    def step(self) -> Tuple[float, np.ndarray, np.ndarray]:
        self.vertices = (
            self.vertices - tf.reduce_mean(self.vertices, axis=0)[None, :]
        )  # centervertices because we have another paramter to control translations

        with tf.GradientTape() as tape:
            vertices_with_grad = tf.constant(self.vertices)
            quaternion_with_grad = tf.constant(self.transform_quaternion)
            translation_with_grad = tf.constant(self.transform_translation)

            tape.watch(vertices_with_grad)
            tape.watch(quaternion_with_grad)
            tape.watch(translation_with_grad)

            vertices_with_grad_centered = (
                vertices_with_grad - tf.reduce_mean(vertices_with_grad, axis=0)[None, :]
            )

            q_normalized = quaternion_with_grad / tf.norm(
                quaternion_with_grad
            )  # that will lead to a gradient that is in the tangeant space
            vertices_with_grad_transformed = (
                qrot(q_normalized, vertices_with_grad_centered) + translation_with_grad
            )

            self.mesh.set_vertices(vertices_with_grad_transformed)

            depth_scale = 1 * self.depthScale
            depth = self.scene.render_depth(
                self.camera,
                depth_scale=depth_scale,
            )
            depth = tf.clip_by_value(depth, 0, self.max_depth)

            diff_image = tf.reduce_sum(
                (depth - tf.constant(self.mesh_image[:, :, None])) ** 2, axis=2
            )
            loss = tf.reduce_sum(diff_image)

            trainable_variables = [
                vertices_with_grad,
                quaternion_with_grad,
                translation_with_grad,
            ]
            vertices_grad, quaternion_grad, translation_grad = tape.gradient(
                loss, trainable_variables
            )

        energy_data = loss.numpy()

        grad_data = vertices_grad

        (
            energy_rigid,
            grad_rigidity,
            _,
        ) = self.rigid_energy.evaluate(self.vertices.numpy())
        energy = energy_data + energy_rigid
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))

        # update v
        grad = grad_data + grad_rigidity

        # update vertices
        step_vertices = mult_and_clamp(
            -grad.numpy(), self.step_factor_vertices, self.step_max_vertices
        )
        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * self.inertia + (1 - self.inertia) * step_vertices
        )
        self.vertices = self.vertices + self.speed_vertices
        # update rotation
        step_quaternion = mult_and_clamp(
            -quaternion_grad.numpy(),
            self.step_factor_quaternion,
            self.step_max_quaternion,
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * self.inertia + (1 - self.inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        # update translation
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )
        step_translation = mult_and_clamp(
            -translation_grad.numpy(),
            self.step_factor_translation,
            self.step_max_translation,
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * self.inertia
            + (1 - self.inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation

        self.iter += 1
        return energy, depth[:, :, 0].numpy(), diff_image.numpy()  # type: ignore


class MeshRGBFitterWithPose:
    """Class to fit a deformable mesh to a color image using Tensorflow."""

    def __init__(
        self,
        vertices: np.ndarray,
        faces: np.ndarray,
        euler_init: np.ndarray,
        translation_init: np.ndarray,
        default_color: np.ndarray,
        default_light_directional: np.ndarray,
        default_light_ambient: float,
        cregu: float = 2000,
        inertia: float = 0.96,
        damping: float = 0.05,
        update_lights: bool = True,
        update_color: bool = True,
    ):
        self.cregu = cregu

        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 0.5
        self.step_factor_quaternion = 0.00006
        self.step_max_quaternion = 0.05
        self.step_factor_translation = 0.00005
        self.step_max_translation = 0.1

        self.default_color = default_color
        self.default_light_directional = default_light_directional
        self.default_light_ambient = default_light_ambient
        self.update_lights = update_lights
        self.update_color = update_color
        self.mesh = ColoredTriMesh(
            faces=faces.copy(),
            vertices=vertices,
            colors=np.zeros((vertices.shape[0], 0)),
        )  # we do a copy to avoid negative stride not support by Tensorflow
        object_center = vertices.mean(axis=0) + translation_init
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([0, 0, 9]) * object_radius

        self.scene = Scene3DTensorflow()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergyTensorflow(self.mesh, vertices, cregu)
        self.vertices_init = tf.constant(copy.copy(vertices))
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)
        self.reset()

    def set_background_color(self, background_color: np.ndarray) -> None:
        self.scene.set_background_color(background_color)

    def set_mesh_transform_init(
        self, euler: np.ndarray, translation: np.ndarray
    ) -> None:
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self) -> None:
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

        self.mesh_color = copy.copy(self.default_color)
        self.light_directional = copy.copy(self.default_light_directional)
        self.light_ambient = self.default_light_ambient

        self.speed_light_directional = np.zeros(self.light_directional.shape)
        self.speed_light_ambient = 0.0
        self.speed_mesh_color = np.zeros(self.mesh_color.shape)

    def set_image(
        self,
        mesh_image: np.ndarray,
        focal: Optional[float] = None,
        distortion: Optional[np.ndarray] = None,
    ) -> None:
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 3
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = CameraTensorflow(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            distortion=distortion,
            width=self.width,
            height=self.height,
        )
        self.iter = 0

    def step(self) -> Tuple[float, np.ndarray, np.ndarray]:
        with tf.GradientTape() as tape:
            vertices_with_grad = tf.constant(self.vertices)
            quaternion_with_grad = tf.constant(self.transform_quaternion)
            translation_with_grad = tf.constant(self.transform_translation)

            light_directional_with_grad = tf.constant(self.light_directional)
            light_ambient_with_grad = tf.constant(
                [self.light_ambient], dtype=np.float64
            )
            mesh_color_with_grad = tf.constant(self.mesh_color)

            tape.watch(vertices_with_grad)
            tape.watch(quaternion_with_grad)
            tape.watch(translation_with_grad)

            tape.watch(light_directional_with_grad)
            tape.watch(light_ambient_with_grad)
            tape.watch(mesh_color_with_grad)

            vertices_with_grad_centered = (
                vertices_with_grad - tf.reduce_mean(vertices_with_grad, axis=0)[None, :]
            )

            q_normalized = quaternion_with_grad / tf.norm(
                quaternion_with_grad
            )  # that will lead to a gradient that is in the tangeant space
            vertices_with_grad_transformed = (
                qrot(q_normalized, vertices_with_grad_centered) + translation_with_grad
            )
            self.mesh.set_vertices(vertices_with_grad_transformed)

            self.scene.set_light(
                light_directional=light_directional_with_grad,
                light_ambient=light_ambient_with_grad,
            )
            self.mesh.set_vertices_colors(
                tf.tile(mesh_color_with_grad[None, :], [self.mesh.nb_vertices, 1])
            )

            image = self.scene.render(self.camera)

            diff_image = tf.reduce_sum(
                (image - tf.constant(self.mesh_image)) ** 2, axis=2
            )
            loss = tf.reduce_sum(diff_image)

            trainable_variables = [
                vertices_with_grad,
                quaternion_with_grad,
                translation_with_grad,
                light_directional_with_grad,
                light_ambient_with_grad,
                mesh_color_with_grad,
            ]
            (
                vertices_grad,
                quaternion_grad,
                translation_grad,
                light_directional_grad,
                light_ambient_grad,
                mesh_color_grad,
            ) = tape.gradient(loss, trainable_variables)

        energy_data = loss.numpy()

        grad_data = vertices_grad

        (
            energy_rigid,
            grad_rigidity,
            approx_hessian_rigidity,
        ) = self.rigid_energy.evaluate(self.vertices)
        energy = energy_data + energy_rigid.numpy()
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))

        # update v
        grad = grad_data + grad_rigidity

        inertia = self.inertia

        # update vertices
        step_vertices = mult_and_clamp(
            -grad.numpy(), self.step_factor_vertices, self.step_max_vertices
        )
        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * inertia + (1 - inertia) * step_vertices
        )
        self.vertices = self.vertices + self.speed_vertices
        # update rotation
        step_quaternion = mult_and_clamp(
            -quaternion_grad, self.step_factor_quaternion, self.step_max_quaternion
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * inertia + (1 - inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        # update translation
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )
        step_translation = mult_and_clamp(
            -translation_grad, self.step_factor_translation, self.step_max_translation
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * inertia + (1 - inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation
        # update directional light
        step = -light_directional_grad * 0.0001
        self.speed_light_directional = (1 - self.damping) * (
            self.speed_light_directional * inertia + (1 - inertia) * step
        )
        self.light_directional = self.light_directional + self.speed_light_directional
        # update ambient light
        step = -light_ambient_grad * 0.0001
        self.speed_light_ambient = (1 - self.damping) * (
            self.speed_light_ambient * inertia + (1 - inertia) * step
        )
        self.light_ambient = self.light_ambient + self.speed_light_ambient.numpy()[0]  # type: ignore
        # update mesh color
        step = -mesh_color_grad * 0.00001
        self.speed_mesh_color = (1 - self.damping) * (
            self.speed_mesh_color * inertia + (1 - inertia) * step
        )
        self.mesh_color = self.mesh_color + self.speed_mesh_color

        self.iter += 1
        return energy, image.numpy(), diff_image.numpy()  # type: ignore



"""Test using rgb_image hand fitting."""

import numpy as np
import tensorflow as tf

from deodr.examples.rgb_image_hand_fitting import run


def test_rgb_image_hand_fitting_pytorch() -> None:
    energies = run(
        dl_library="pytorch",
        plot_curves=False,
        display=False,
        save_images=False,
        max_iter=50,
    )

    possible_results = [
        2103.665850588039,  # github action ubuntu python 3.7
        2104.9656991756697,  # github action ubuntu python 3.9
        2100.0239709048583,
        2132.9307950405196,
        2106.5436357944604,
        2117.9946156293213,  # google colab
        2112.3014481160876,  # github action windows python 3.8
        2108.0875835193865,  # github action ubuntu python 3.7
    ]

    assert np.any(np.abs(np.array(possible_results) - energies[49]) < 1e-5)


def test_rgb_image_hand_fitting_numpy() -> None:
    energies = run(
        dl_library="none",
        plot_curves=False,
        display=False,
        save_images=False,
        max_iter=50,
    )
    # getting different result on python 3.6 or 3.7 in the github action, not sure why
    possible_results = [
        2122.8322696714026,  # python 3.7 on github action
        2107.850380422819,
        2107.850380422819,  # google colab Intel(R) Xeon(R) CPU @ 2.20GHz
        2113.7013184079137,  # python 3.7 on windows on github action
    ]

    assert np.any(np.abs(np.array(possible_results) - energies[49]) < 1e-5)


def test_rgb_image_hand_fitting_tensorflow() -> None:
    tf.config.set_visible_devices(
        [], "GPU"
    )  # Running on CPU to get deterministic results

    energies = run(
        dl_library="tensorflow",
        plot_curves=False,
        display=False,
        save_images=False,
        max_iter=50,
    )
    possible_results = [
        2112.9566220857746,
        2115.9320061795634,
        2107.962374538259,
        2115.9974345976066,
        2115.791145850562,
        2114.564762503404,
        2102.9991446481354,
        2167.4449854352574,
    ]

    assert np.any(np.abs(np.array(possible_results) - energies[49]) < 1e-5)


if __name__ == "__main__":
    test_rgb_image_hand_fitting_tensorflow()
    test_rgb_image_hand_fitting_pytorch()

    test_rgb_image_hand_fitting_numpy()



"""Small tools for tensorflow."""

import numpy as np
import scipy.sparse
import tensorflow as tf


def scipy_sparse_matrix_to_tensorflow(x: scipy.sparse.spmatrix) -> tf.SparseTensor:
    coo = x.tocoo()
    indices = np.mat([coo.row, coo.col]).transpose()
    return tf.SparseTensor(indices, coo.data, coo.shape)



"""Example with fitting a colored hand mesh model to multiple images."""
import datetime
import glob
import json
import os
import time

import cv2
from imageio import imread, imwrite
import matplotlib.pyplot as plt
import numpy as np
from typing_extensions import Literal

import deodr
from deodr import read_obj
from deodr.mesh_fitter import MeshRGBFitterWithPoseMultiFrame

dl_library_type = Literal["pytorch", "tensorflow", "none"]


def run(
    dl_library: dl_library_type = "pytorch",
    plot_curves: bool = False,
    save_images: bool = False,
    display: bool = True,
    max_iter: int = 400,
) -> None:
    file_folder = os.path.dirname(__file__)
    hand_images = [
        imread(file).astype(np.double) / 255
        for file in glob.glob(os.path.join(deodr.data_path, "./hand_multiview/*.jpg"))
    ]
    nb_frames = len(hand_images)

    obj_file = os.path.join(deodr.data_path, "hand.obj")
    faces, vertices = read_obj(obj_file)

    default_color = np.array([0.4, 0.3, 0.25]) * 1.5
    default_light_directional = -np.array([0.1, 0.5, 0.4])
    default_light_ambient = 0.6
    # default_light = {'directional':np.array([0.0,0.0,0.0]),'ambient':np.array([0.6])}

    euler_init = np.row_stack(
        [np.array([0, yrot, 0]) for yrot in np.linspace(-0.5, 0.5, 3)]
    )

    vertices = vertices - np.mean(vertices, axis=0)
    t_init = np.array([0, -0.2, 0.2])
    translation_init = np.tile(t_init[None, :], [nb_frames, 1])
    # centering vertices

    hand_fitter = MeshRGBFitterWithPoseMultiFrame(
        vertices,
        faces,
        default_color=default_color,
        default_light_directional=default_light_directional,
        default_light_ambient=default_light_ambient,
        update_lights=True,
        update_color=True,
        euler_init=euler_init,
        translation_init=translation_init,
        cregu=2000,
    )
    #  handFitter = MeshRGBFitter(vertices,faces,default_color,default_light,
    # update_lights =  True, update_color= True,cregu=1000)

    hand_fitter.reset()

    max_iter = 300

    hand_image = hand_images[0]
    background_color = np.median(
        np.row_stack(
            (
                hand_image[:10, :10, :].reshape(-1, 3),
                hand_image[-10:, :10, :].reshape(-1, 3),
                hand_image[-10:, -10:, :].reshape(-1, 3),
                hand_image[:10, -10:, :].reshape(-1, 3),
            )
        ),
        axis=0,
    )
    background_color = np.array([0, 0, 0])
    hand_fitter.set_images(hand_images)
    hand_fitter.set_background_color(background_color)
    energies = []
    durations = []
    start = time.time()

    iter_folder = os.path.join(file_folder, "./iterations/depth")
    if not os.path.exists(iter_folder):
        os.makedirs(iter_folder)

    for niter in range(max_iter):
        energy, image, diff_image = hand_fitter.step(check_gradient=False)
        energies.append(energy)
        durations.append(time.time() - start)
        if display or save_images:
            combined_image = np.column_stack(
                (
                    np.row_stack(hand_images),
                    image,
                    np.tile(np.minimum(diff_image, 1)[:, :, None], (1, 1, 3)),
                )
            )
        if display:
            cv2.imshow(
                "animation",
                cv2.resize(combined_image[:, :, ::-1], None, fx=1, fy=1),
            )
        if save_images:
            imwrite(
                os.path.join(iter_folder, f"hand_iter_{niter}.png"),
                (combined_image * 255).astype(np.uint8),
            )
        cv2.waitKey(1)

    # save convergence curve
    with open(
        os.path.join(
            iter_folder,
            f'rgb_image_fitting_result_{str(datetime.datetime.now()).replace(":", "_")}.json',
        ),
        "w",
    ) as f:
        json.dump(
            {
                "label": f"{dl_library} {datetime.datetime.now()}",
                "durations": durations,
                "energies": energies,
            },
            f,
            indent=4,
        )

    # compare with previous runs
    if plot_curves:
        plt.figure()
        for file in glob.glob(
            os.path.join(iter_folder, "rgb_image_fitting_result_*.json")
        ):
            with open(file, "r") as fp:
                json_data = json.load(fp)
                plt.plot(
                    json_data["durations"],
                    json_data["energies"],
                    label=json_data["label"],
                )
                plt.xlabel("duration in seconds")
                plt.ylabel("energies")
        plt.legend()
        plt.figure()
        for file in glob.glob(
            os.path.join(iter_folder, "rgb_image_fitting_result_*.json")
        ):
            with open(file, "r") as fp:
                json_data = json.load(fp)
                plt.plot(json_data["energies"], label=json_data["label"])
                plt.xlabel("iterations")
                plt.ylabel("energies")
        plt.legend()
        plt.show()


if __name__ == "__main__":
    display = True
    save_images = False
    run(dl_library="none", plot_curves=True, display=display, save_images=save_images)



"""Example of interactive 3D mesh visualization using DEODR and OpenCV."""

import argparse
import os
import pickle
import time
from typing import Any, Callable, Dict, Optional, Tuple, Union

import cv2
import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial.transform import Rotation
import trimesh
from typing_extensions import Literal

import deodr
from deodr import differentiable_renderer
from deodr.differentiable_renderer import Camera
from deodr.opengl.moderngl import OffscreenRenderer  # type: ignore
from deodr.triangulated_mesh import ColoredTriMesh

InteractorModeType = Literal["camera_centered", "object_centered_trackball"]


class Interactor:
    """Class that implements various mouse interaction with the 3D scene."""

    def __init__(
        self,
        camera: Camera,
        mode: InteractorModeType = "object_centered_trackball",
        object_center: Optional[np.ndarray] = None,
        rotation_speed: float = 0.003,
        z_translation_speed: float = 0.05,
        xy_translation_speed: float = 0.01,
    ):
        self.left_is_down = False
        self.right_is_down = False
        self.middle_is_down = False
        self.mode = mode
        if mode == "object_centered_trackball":
            assert object_center is not None
            assert object_center.shape == (3,)
        elif object_center is None:
            object_center = np.array([0, 0, 0])

        self.object_center = object_center
        self.rotation_speed = rotation_speed
        self.z_translation_speed = z_translation_speed
        self.xy_translation_speed = xy_translation_speed
        self.camera = camera

    def toggle_mode(self) -> None:
        if self.mode == "object_centered_trackball":
            self.mode = "camera_centered"
        else:
            self.mode = "object_centered_trackball"
        print(f"trackball mode = {self.mode}")

    def rotate(
        self,
        rot_vec: np.ndarray,
    ) -> None:
        assert np.array(rot_vec).shape == (3,)
        rotation = Rotation.from_rotvec(np.array(rot_vec))
        if self.mode == "camera_centered":
            self.camera.extrinsic = rotation.as_matrix().dot(self.camera.extrinsic)
        else:
            n_rotation = rotation.as_matrix().dot(self.camera.extrinsic[:, :3])
            nt = (
                self.camera.extrinsic[:, :3].dot(self.object_center)
                + self.camera.extrinsic[:, 3]
                - n_rotation.dot(self.object_center)
            )
            self.camera.extrinsic = np.column_stack((n_rotation, nt))

    def mouse_callback(
        self, event: int, x: int, y: int, flags: int, param: Any
    ) -> None:
        if event == 0 and flags == 0:
            return
        if event == cv2.EVENT_LBUTTONDOWN:
            self.left_is_down = True
            self.x_last = x
            self.y_last = y
        # check to see if the left mouse button was released
        if event == cv2.EVENT_LBUTTONUP:
            self.left_is_down = False
        if event == cv2.EVENT_RBUTTONDOWN:
            self.right_is_down = True
            self.x_last = x
            self.y_last = y
            # check to see if the left mouse button was released
        if event == cv2.EVENT_RBUTTONUP:
            self.right_is_down = False

        if event == cv2.EVENT_MBUTTONDOWN:
            self.middle_is_down = True
            self.x_last = x
            self.y_last = y
        # check to see if the left mouse button was released
        if event == cv2.EVENT_MBUTTONUP:
            self.middle_is_down = False

        self.ctrl_is_down = flags & cv2.EVENT_FLAG_CTRLKEY
        self.shift_is_down = flags & cv2.EVENT_FLAG_SHIFTKEY

        if self.left_is_down and not (self.ctrl_is_down):
            if self.mode == "camera_centered":
                rot_vec = np.array(
                    [
                        -0.3 * self.rotation_speed * (y - self.y_last),
                        0.3 * self.rotation_speed * (x - self.x_last),
                        0,
                    ]
                )
                self.rotate(rot_vec)

                # assert np.allclose(center_in_camera, self.camera.world_to_camera(self.object_center))
                self.x_last = x
                self.y_last = y

            elif self.mode == "object_centered_trackball":
                self.rotate(
                    np.array(
                        [
                            self.rotation_speed * (y - self.y_last),
                            -self.rotation_speed * (x - self.x_last),
                            0,
                        ]
                    )
                )

                self.x_last = x
                self.y_last = y
            else:
                raise (BaseException(f"unknown camera mode {self.mode}"))

        if self.right_is_down and self.shift_is_down:
            delta_y = self.y_last - y
            ratio = np.power(2, delta_y / 20)
            self.camera.intrinsic[0, 0] = self.camera.intrinsic[0, 0] * ratio
            self.camera.intrinsic[1, 1] = self.camera.intrinsic[1, 1] * ratio
            self.x_last = x
            self.y_last = y

        if self.right_is_down and not (self.ctrl_is_down):
            if self.mode in ["camera_centered", "object_centered_trackball"]:
                if np.abs(self.y_last - y) >= np.abs(self.x_last - x):
                    self.camera.extrinsic[2, 3] += self.z_translation_speed * (
                        self.y_last - y
                    )
                else:
                    self.rotate(
                        np.array(
                            [
                                0,
                                0,
                                -self.rotation_speed * (self.x_last - x),
                            ]
                        )
                    )
                self.x_last = x
                self.y_last = y

            else:
                raise (BaseException(f"unknown camera mode {self.mode}"))

        if self.middle_is_down or (self.left_is_down and self.ctrl_is_down):
            # translation

            object_depth = (
                self.camera.extrinsic[2, :3].dot(self.object_center)
                + self.camera.extrinsic[2, 3]
            )

            tx = self.xy_translation_speed * object_depth * (x - self.x_last)
            ty = self.xy_translation_speed * object_depth * (y - self.y_last)

            self.object_center -= (
                self.camera.extrinsic[0, :3] * tx + self.camera.extrinsic[1, :3] * ty
            )
            self.camera.extrinsic[0, 3] += tx
            self.camera.extrinsic[1, 3] += ty
            self.x_last = x
            self.y_last = y

    def print_help(self) -> None:
        help_str = "" + "Mouse:\n"
        if self.mode == "object_centered_trackball":
            help_str += (
                "mouse left + vertical motion: rotate object along camera x axis\n"
            )
            help_str += (
                "mouse left + horizontal motion: rotate object along camera y axis\n"
            )
            help_str += (
                "mouse right + vertical motion: translate object along camera z axis\n"
            )
            help_str += (
                "mouse right + horizontal motion: rotate object along camera z axis\n"
            )
            help_str += "CTRL + mouse left + vertical motion: translate object along camera y axis\n"
            help_str += "CTRL + mouse left + horizontal motion: translate object along camera x axis\n"

        else:
            help_str += (
                "mouse right + vertical motion: translate camera along its z axis\n"
            )
            help_str += (
                "mouse right + horizontal motion: rotate camera along its z axis\n"
            )
            help_str += "mouse left + vertical motion: rotate camera along its x axis\n"
            help_str += (
                "mouse left + horizontal motion: rotate camera along its y axis\n"
            )
            help_str += "CTRL + mouse left + vertical motion: translate camera along its y axis\n"
            help_str += "CTRL + mouse left + horizontal motion: translate camera along its x axis\n"
        help_str += (
            "SHIFT + mouse left + vertical motion: change the camera field of view\n"
        )
        print(help_str)


class Viewer:
    def __init__(
        self,
        file_or_mesh: Union[str, ColoredTriMesh],
        display_texture_map: bool = True,
        width: int = 640,
        height: int = 480,
        display_fps: bool = True,
        title: Optional[str] = None,
        use_moderngl: bool = False,
        light_directional: Tuple[float, float, float] = (0, 0, -0.5),
        light_ambient: float = 0.5,
        background_color: Tuple[float, float, float] = (1, 1, 1),
        use_antialiasing: bool = True,
        use_light: bool = True,
        fps_exp_average_decay: float = 0.1,
        horizontal_fov: float = 60,
        video_pattern: str = "deodr_viewer_recording{id}.avi",
        video_format: str = "MJPG",
    ):
        self.title = title
        self.scene = differentiable_renderer.Scene3D(sigma=1)
        self.set_mesh(file_or_mesh)
        self.windowname = f"DEODR mesh viewer:{self.title}"

        self.width = width
        self.height = height
        self.display_fps = display_fps
        self.use_moderngl = use_moderngl
        self.use_antialiasing = use_antialiasing
        self.use_light = use_light
        self.fps_exp_average_decay = fps_exp_average_decay
        self.last_time: Optional[float] = None
        self.horizontal_fov = horizontal_fov
        self.video_writer: Optional[cv2.VideoWriter] = None
        self.recording = False
        self.video_pattern = video_pattern
        self.video_format = video_format
        self.fps: float = 0

        if display_texture_map:
            self.display_texture_map()

        self.set_background_color(background_color)
        self.set_light(light_directional, light_ambient)
        self.recenter_camera()

        self.offscreen_renderer: Optional[OffscreenRenderer] = None
        if use_moderngl:
            self.setup_moderngl()

        self.register_keys()

    def set_light(
        self,
        light_directional: Union[Tuple[float, float, float], np.ndarray],
        light_ambient: float,
    ) -> None:
        self.light_directional = np.array(light_directional)
        self.light_ambient = light_ambient
        self.scene.set_light(
            light_directional=self.light_directional, light_ambient=light_ambient
        )

    def setup_moderngl(self) -> None:
        self.offscreen_renderer = OffscreenRenderer()
        assert self.scene.mesh is not None
        self.scene.mesh.compute_vertex_normals()
        self.offscreen_renderer.set_scene(self.scene)

    def set_background_color(
        self, background_color: Tuple[float, float, float]
    ) -> None:
        self.scene.set_background_color(background_color)

    def display_texture_map(self) -> None:
        if self.mesh.textured:
            ax = plt.subplot(111)
            self.mesh.plot_uv_map(ax)

    def set_mesh(self, file_or_mesh: Union[str, ColoredTriMesh]) -> None:
        if isinstance(file_or_mesh, str):
            if self.title is None:
                self.title = file_or_mesh
            mesh_trimesh = trimesh.load(file_or_mesh)
            self.mesh = ColoredTriMesh.from_trimesh(mesh_trimesh)
        elif isinstance(file_or_mesh, trimesh.base.Trimesh):
            mesh_trimesh = file_or_mesh
            self.mesh = ColoredTriMesh.from_trimesh(mesh_trimesh)
            if self.title is None:
                self.title = "unknown"
        elif isinstance(file_or_mesh, ColoredTriMesh):
            self.mesh = file_or_mesh
            if self.title is None:
                self.title = "unknown"
        else:
            raise (
                TypeError(
                    f"unknown type {type(file_or_mesh)} for input obj_file_or_trimesh,"
                    " can be string or trimesh.base.Trimesh"
                )
            )
        self.object_center = 0.5 * (
            self.mesh.vertices.max(axis=0) + self.mesh.vertices.min(axis=0)
        )
        self.object_radius = np.max(
            self.mesh.vertices.max(axis=0) - self.mesh.vertices.min(axis=0)
        )
        self.scene.set_mesh(self.mesh)

    def recenter_camera(self) -> None:
        camera_center = self.object_center + np.array([0, 0, 3]) * self.object_radius
        rotation = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        translation = -rotation.T.dot(camera_center)
        extrinsic = np.column_stack((rotation, translation))
        focal = 0.5 * self.width / np.tan(0.5 * self.horizontal_fov * np.pi / 180)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )

        distortion = [0, 0, 0, 0, 0]
        self.camera = differentiable_renderer.Camera(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            width=self.width,
            height=self.height,
            distortion=distortion,
        )

        self.interactor = Interactor(
            camera=self.camera,
            object_center=self.object_center,
            z_translation_speed=0.01 * self.object_radius,
            xy_translation_speed=3e-4,
        )

    def start(self, print_help: bool = True, loop: bool = True) -> None:
        """Open the window and start the loop if loop true."""
        if print_help:
            self.print_help()
        self.fps = 0
        cv2.namedWindow(self.windowname, cv2.WINDOW_NORMAL)
        cv2.resizeWindow(self.windowname, self.width, self.height)
        cv2.setMouseCallback(self.windowname, self.interactor.mouse_callback)  # type: ignore
        if loop:
            while cv2.getWindowProperty(self.windowname, 0) >= 0:
                self.refresh()

    def update_fps(self) -> None:
        new_time = time.perf_counter()
        if self.last_time is None:
            self.fps = 0
        elif self.fps == 0:
            self.fps = 1 / (new_time - self.last_time)
        else:
            new_fps = 1 / (new_time - self.last_time)
            self.fps = (
                1 - self.fps_exp_average_decay
            ) * self.fps + self.fps_exp_average_decay * new_fps
        self.last_time = new_time

    def refresh(self) -> None:
        self.width, self.height = cv2.getWindowImageRect(self.windowname)[2:]
        self.resize_camera()

        if self.use_moderngl:
            assert self.offscreen_renderer is not None
            image = self.offscreen_renderer.render(self.camera)
        else:
            image = (self.scene.render(self.interactor.camera) * 255).astype(np.uint8)

        bgr_image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        if self.recording:
            assert self.video_writer is not None
            self.video_writer.write(bgr_image.astype(np.uint8))

        self.update_fps()
        if self.recording:
            cv2.circle(
                bgr_image,
                (image.shape[1] - 20, image.shape[0] - 20),
                8,
                (0, 0, 255),
                cv2.FILLED,
            )
        if self.display_fps:
            self.print_fps(bgr_image, self.fps)

        cv2.imshow(self.windowname, bgr_image)

        key = cv2.waitKey(1)
        if key > 0:
            self.process_key(key)

    def resize_camera(self) -> None:
        ratio = self.width / self.camera.width

        intrinsic = np.array(
            [
                [self.camera.intrinsic[0, 0] * ratio, 0, self.width / 2],
                [0, self.camera.intrinsic[1, 1] * ratio, self.height / 2],
                [0, 0, 1],
            ]
        )
        self.camera.intrinsic = intrinsic
        self.camera.width = self.width
        self.camera.height = self.height

    def print_fps(self, image: np.ndarray, fps: float) -> None:
        font = cv2.FONT_HERSHEY_SIMPLEX
        bottom_left_corner_of_text = (20, image.shape[0] - 20)
        font_scale = 1
        font_color = (0, 0, 255)
        thickness = 2
        cv2.putText(
            image,
            "fps:%0.1f" % fps,
            bottom_left_corner_of_text,
            font,
            font_scale,
            font_color,
            thickness,
        )

    def print_help(self) -> None:
        """Print the help message."""
        help_str = "" + "-----------------\n"
        help_str += "DEODR Mesh Viewer\n"
        help_str += "-----------------\n"
        help_str += "Keys:\n"
        for key, func in self.keys_map.items():
            help_str += f"{key}: {func.__doc__}\n"
        print(help_str)
        self.interactor.print_help()

    def toggle_renderer(self) -> None:
        """Toggle the renderer between DEODR cpu rendering and moderngl."""
        self.use_moderngl = not (self.use_moderngl)
        print(f"use_moderngl = { self.use_moderngl}")

        if self.use_moderngl and self.offscreen_renderer is None:
            self.setup_moderngl()

    def toggle_perspective_texture_mapping(self) -> None:
        """Toggle between linear texture mapping and perspective correct texture mapping."""
        if self.use_moderngl:
            print("can only use perspective correct mapping  when using moderngl")
        else:
            self.scene.perspective_correct = not (self.scene.perspective_correct)
            print(f"perspective_correct = {self.scene.perspective_correct}")

    def toggle_lights(self) -> None:
        """Toggle between uniform lighting vs directional + ambient."""
        self.use_light = not (self.use_light)
        print(f"use_light = { self.use_light}")

        if self.use_light:
            if self.use_moderngl:
                assert self.offscreen_renderer is not None
                self.offscreen_renderer.set_light(
                    light_directional=np.array(self.light_directional),
                    light_ambient=self.light_ambient,
                )
            else:
                self.scene.set_light(
                    light_directional=np.array(self.light_directional),
                    light_ambient=self.light_ambient,
                )
        elif self.use_moderngl:
            assert self.offscreen_renderer is not None
            self.offscreen_renderer.set_light(
                light_directional=np.ndarray((0, 0, 0)),
                light_ambient=1.0,
            )
        else:
            self.scene.set_light(light_directional=(0, 0, 0), light_ambient=1.0)

    def toggle_edge_overdraw_antialiasing(self) -> None:
        """Toggle edge overdraw anti-aliasing (DEODR rendering only)."""
        if self.use_moderngl:
            print("no anti-aliasing available when using moderngl")
        else:
            self.use_antialiasing = not (self.use_antialiasing)
            print(f"use_antialiasing = {self.use_antialiasing}")
            self.scene.sigma = 1.0 if self.use_antialiasing else 0.0

    def pickle_scene_and_cameras(self) -> None:
        """Save scene and camera in a pickle file."""
        filename = os.path.abspath("scene.pickle")
        # save scene and camera in pickle file
        with open(filename, "wb") as file:
            # dump information to the file
            pickle.dump(self.scene, file)
        print(f"saved scene in {filename}")

        filename = os.path.abspath("camera.pickle")
        print(f"save scene in {filename}")
        with open(filename, "wb") as file:
            # dump information to the file
            pickle.dump(self.camera, file)
        print(f"saved camera in {filename}")

    def toggle_interactor_mode(self) -> None:
        """Change the camera interactor mode."""
        self.interactor.toggle_mode()
        self.interactor.print_help()

    def toggle_video_recording(self) -> None:
        """Start and stop video recording."""
        if not self.recording:
            id = 0
            while os.path.exists(self.video_pattern.format(**dict(id=id))):
                id += 1
            filename = self.video_pattern.format(**dict(id=id))

            self.video_writer = cv2.VideoWriter(
                filename,
                cv2.VideoWriter_fourcc(*self.video_format),  # type: ignore
                30,
                (self.width, self.height),
            )
            self.recording = True
        else:
            assert self.video_writer is not None
            self.video_writer.release()
            self.recording = False

    def register_keys(self) -> None:
        self.keys_map: Dict[str, Callable[[], None]] = {}
        self.register_key("h", self.print_help)
        self.register_key("r", self.toggle_renderer)
        self.register_key("p", self.toggle_perspective_texture_mapping)
        self.register_key("l", self.toggle_lights)
        self.register_key("a", self.toggle_edge_overdraw_antialiasing)
        self.register_key("d", self.pickle_scene_and_cameras)
        self.register_key("s", self.toggle_video_recording)
        self.register_key("t", self.toggle_interactor_mode)

    def register_key(self, key: str, func: Callable[[], None]) -> None:
        self.keys_map[key] = func

    def process_key(self, key: int) -> None:
        chr_key = chr(key)
        if chr_key in self.keys_map:
            self.keys_map[chr(key)]()
        else:
            print(f"no function registered for key {chr_key}")


def run() -> None:
    obj_file = os.path.join(deodr.data_path, "duck.obj")
    Viewer(obj_file, use_moderngl=False).start()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="mesh_viewer", usage="%(prog)s [options]")
    duck_file = os.path.join(deodr.data_path, "duck.obj")
    parser.add_argument("mesh_file", type=str, nargs="?", default=duck_file)
    args = parser.parse_args()
    mesh_file = args.mesh_file
    Viewer(mesh_file, use_moderngl=True).start()



"""Examples with 3D mesh rendering using various backend and comparison with deodr."""

import os
from typing import Tuple

import imageio.v3 as imageio
import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial.transform import Rotation
import trimesh

import deodr
from deodr import differentiable_renderer
from deodr.differentiable_renderer import Camera, Scene3D
from deodr.triangulated_mesh import ColoredTriMesh


def default_scene(
    obj_file: str,
    width: int = 640,
    height: int = 480,
    use_distortion: bool = True,
    integer_pixel_centers: bool = True,
) -> Tuple[Scene3D, Camera]:
    mesh_trimesh = trimesh.load(obj_file)

    mesh = ColoredTriMesh.from_trimesh(mesh_trimesh)

    # rot = Rotation.from_euler("xyz", [180, 0, 0], degrees=True).as_matrix()
    rot = Rotation.from_euler("xyz", [180, 0, 0], degrees=True).as_matrix()

    camera = differentiable_renderer.default_camera(
        width, height, 80, mesh.vertices, rot
    )
    if use_distortion:
        camera.distortion = np.array([-0.5, 0.5, 0, 0, 0])

    bg_color = np.array((0.8, 0.8, 0.8))
    scene = differentiable_renderer.Scene3D(integer_pixel_centers=integer_pixel_centers)
    light_ambient = 0
    light_directional = 0.3 * np.array([1, -1, 0])
    scene.set_light(light_directional=light_directional, light_ambient=light_ambient)
    scene.set_mesh(mesh)
    scene.set_background_color(bg_color)
    return scene, camera


def example_rgb(
    display: bool = True, save_image: bool = False, width: int = 640, height: int = 480
) -> np.ndarray:
    obj_file = os.path.join(deodr.data_path, "duck.obj")
    scene, camera = default_scene(obj_file, width=width, height=height)
    image = scene.render(camera)
    if save_image:
        image_file = os.path.abspath(os.path.join(deodr.data_path, "test/duck.png"))
        os.makedirs(os.path.dirname(image_file), exist_ok=True)
        image_uint8 = (image * 255).astype(np.uint8)
        imageio.imwrite(image_file, image_uint8)
    if display:
        plt.figure()
        plt.title("deodr rendering")
        plt.imshow(image)
    return image


def normalize_unit_cube(v: np.ndarray) -> np.ndarray:
    if v.ndim == 3 and v.shape[2] < 3:
        nv = np.zeros((v.shape[0], v.shape[1], 3))
        nv[:, :, : v.shape[2]] = v
    else:
        nv = v
    return (nv - nv.min()) / (nv.max() - nv.min())


def example_channels(
    display: bool = True, save_image: bool = False, width: int = 640, height: int = 480
) -> None:
    obj_file = os.path.join(deodr.data_path, "duck.obj")
    scene, camera = default_scene(obj_file, width=width, height=height)

    scene.sigma = 0

    channels = scene.render_deferred(camera)
    if display:
        plt.figure()
        for i, (name, v) in enumerate(channels.items()):
            ax = plt.subplot(2, 4, i + 1)
            ax.set_title(name)
            ax.imshow(normalize_unit_cube(v))

    if save_image:
        for name, v in channels.items():
            image_file = os.path.abspath(
                os.path.join(deodr.data_path, f"test/duck_{name}.png")
            )
            os.makedirs(os.path.dirname(image_file), exist_ok=True)
            image_uint8 = (normalize_unit_cube(v) * 255).astype(np.uint8)
            imageio.imwrite(image_file, image_uint8)


def example_pyrender(
    display: bool = True, save_image: bool = False, width: int = 640, height: int = 480
) -> None:
    import deodr.opengl.pyrender

    obj_file = os.path.join(deodr.data_path, "duck.obj")
    scene, camera = default_scene(
        obj_file, use_distortion=False, width=width, height=height
    )
    scene.sigma = 0  # removing edge overdraw antialiasing
    if display:
        image_no_antialiasing = scene.render(camera)
        image_pyrender, depth = deodr.opengl.pyrender.render(scene, camera)
        plt.figure()
        ax = plt.subplot(1, 3, 1)
        ax.set_title("deodr no antialiasing")
        ax.imshow(image_no_antialiasing)

        ax = plt.subplot(1, 3, 2)
        ax.set_title("pyrender")
        ax.imshow(image_pyrender)

        ax = plt.subplot(1, 3, 3)
        ax.set_title("difference")
        ax.imshow(
            np.abs(image_no_antialiasing - image_pyrender.astype(np.float64) / 255)
        )


def example_moderngl(display: bool = True, width: int = 640, height: int = 480) -> None:
    from deodr.opengl.moderngl import OffscreenRenderer  # type: ignore

    obj_file = os.path.join(deodr.data_path, "duck.obj")
    for integer_pixel_centers in [True, False]:
        scene, camera = default_scene(
            obj_file,
            width=width,
            height=height,
            integer_pixel_centers=integer_pixel_centers,
        )
        scene.sigma = 0  # removing edge overdraw antialiasing
        # adding some perturbation to get better test
        camera.extrinsic[1, 2] = camera.extrinsic[1, 2] + 0.1
        camera.extrinsic[1, 1] = camera.extrinsic[1, 1] * 0.9

        image_no_antialiasing = scene.render(camera)
        moderngl_renderer = OffscreenRenderer()
        moderngl_renderer.set_scene(scene)
        image_moderngl = moderngl_renderer.render(camera)
        diff = np.abs(
            image_no_antialiasing.astype(np.float64) * 255
            - image_moderngl.astype(np.float64)
        )
        if display:
            plt.figure()
            ax = plt.subplot(1, 3, 1)
            ax.set_title("deodr no antialiasing")
            ax.imshow(image_no_antialiasing)

            ax = plt.subplot(1, 3, 2)
            ax.set_title("moderngl")
            ax.imshow(image_moderngl)

            ax = plt.subplot(1, 3, 3)
            ax.set_title("difference")
            ax.imshow(np.clip(10 * diff / 255, 0, 1))
            plt.show()

    assert np.sum(np.any(np.abs(diff) > 15, axis=2)) <= 3


if __name__ == "__main__":
    example_moderngl(display=True)
    example_rgb(save_image=False)
    example_channels(save_image=False)
    example_pyrender()
    plt.show()
